package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.BranchInstruction;
import org.gjt.jclasslib.bytecode.ImmediateByteInstruction;
import org.gjt.jclasslib.bytecode.ImmediateShortInstruction;
import org.gjt.jclasslib.bytecode.IncrementInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.bytecode.SimpleInstruction;
import org.gjt.jclasslib.structures.CPInfo;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;
import org.gjt.jclasslib.structures.constants.ConstantMethodrefInfo;
import org.gjt.jclasslib.structures.constants.ConstantNameAndTypeInfo;

public class MethodWrapper extends PDSMethod
{

	private String formattedName;
	private ClassFile classFile;
	private PDSInfo info;
	private MethodInfo method;
	/**
	 * See {@link #getNewVarSlot()}.
	 */
	private int varSlot= 1;
	/**
	 * See {@link #getNewLabelOffset()}.
	 */
	private int labelOffset = 1;
	
	private int maxLocalStack;
	
	public MethodWrapper(MethodInfo method, PDSInfo pdsInfo)
		throws InvalidByteCodeException, IOException 
	{
		super(method.getClassFile(), method, pdsInfo);
		this.classFile = method.getClassFile();
		this.method = method;
		formattedName = "initialWrapperFor_" + PDSMethod.formatMethodName
			(new String[] { classFile.getThisClassName(), method.getName(),
					method.getDescriptor() });
		this.info = pdsInfo;
	}

	public String toRemopla() throws InvalidByteCodeException 
	{
		StringBuilder out = new StringBuilder(toRemoplaHead()).append("\n{\n");
		List<PDSInstOld> instructions = isStatic() ? getStaticWrapperInstructions()
				: getInstanceWrapperInstructions();
				
		for (int i = 0; i < maxLocalStack; i++) {
			out.append("int s").append(i).append(";\n");
		}
		
		for (int i = 0; i < varSlot; i++) {
			out.append("int v").append(i).append(";\n");
		}

		int offset = labelOffset + 1;
		for (PDSInstOld i : instructions) {
			AbstractInstruction inst = i.getInstruction();
			if (inst.getOffset() == 0) {
				inst.setOffset(offset++);
			}
			out.append(i.toRemopla()).append("\n");
		}
		
		return out.append("\n}\n").toString();
	}
	
	private String getDescriptor(ClassFile classFile, int constructorIndex)
		throws InvalidByteCodeException
	{
		ConstantMethodrefInfo info = 
			(ConstantMethodrefInfo)classFile.getConstantPoolEntry(constructorIndex, 
					ConstantMethodrefInfo.class);
		return info.getNameAndTypeInfo().getDescriptor();
	}
	
	private void addPushUndefinedParams(List<PDSInstOld> instructions,  
			String descriptor,	int paramCount, int maxLocalStack) throws InvalidByteCodeException
	{
		for (int i = 0; i < paramCount; i++) {
			addPushUndefInstruction(instructions, maxLocalStack);
			if (PDSMethod.isArray(descriptor, i)) {
				addInstruction(instructions, new ImmediateByteInstruction
					(Opcodes.OPCODE_ANEWARRAY, true, Opcodes.NEWARRAY_T_INT), maxLocalStack);
 				initializeArrayToUndefined(instructions, maxLocalStack);
			}
		}
	}
	
	private void addInstruction(List<PDSInstOld> instructions, AbstractInstruction inst, 
			int maxLocalStack) throws InvalidByteCodeException
	{
		instructions.add(new PDSInstOld(classFile, formattedName, inst, maxLocalStack, getInfo()));
	}
	
	/**
	 * Assumes the array is the current  top element on the operand stack.
	 * @param instructions
	 * @throws InvalidByteCodeException
	 */
	private void initializeArrayToUndefined(List<PDSInstOld> instructions, int maxLocalStack) 
		throws InvalidByteCodeException
	{
		int arraySlot = getNewVarSlot();
		int indexSlot = getNewVarSlot();
		int loopGuardLabelOffset = getNewLabelOffset();
		int loopStartLabelOffset = getNewLabelOffset();
		
		// store array in its var slot
		addInstruction(instructions,
				new ImmediateByteInstruction(Opcodes.OPCODE_ASTORE, false, arraySlot),
				maxLocalStack);

		// load 0 constant
		addInstruction(instructions, new SimpleInstruction(Opcodes.OPCODE_ILOAD_0),
				maxLocalStack);
		
		// store it in index slot
		addInstruction(instructions,
				new ImmediateByteInstruction(Opcodes.OPCODE_ISTORE, false, indexSlot),
				maxLocalStack);
		
		// goto loop guard
		addInstruction(instructions,
				new BranchInstruction(Opcodes.OPCODE_GOTO, loopGuardLabelOffset),
				maxLocalStack);
		
		// load array
		AbstractInstruction instruction =
			new ImmediateByteInstruction(Opcodes.OPCODE_ALOAD, false, arraySlot);
		instruction.setOffset(loopStartLabelOffset);
		addInstruction(instructions, instruction, maxLocalStack);
		
		// load index
		addInstruction(instructions, 
				new ImmediateByteInstruction(Opcodes.OPCODE_ILOAD, false, indexSlot),
				maxLocalStack);

		// push undefined parameter
		addPushUndefInstruction(instructions, maxLocalStack);
		
		// store in array at index
		addInstruction(instructions, new SimpleInstruction(Opcodes.OPCODE_IASTORE),
				maxLocalStack);
		
		// increase index by 1
		addInstruction(instructions, 
				new IncrementInstruction(Opcodes.OPCODE_IINC, false,	indexSlot, 1),
				maxLocalStack);
		
		// load array
		addInstruction(instructions,
				new ImmediateByteInstruction(Opcodes.OPCODE_ALOAD, false, arraySlot),
				maxLocalStack);
		
		// load array length
		addInstruction(instructions, 
				new SimpleInstruction(Opcodes.OPCODE_ARRAYLENGTH),
				maxLocalStack);
		
		instruction = new BranchInstruction(Opcodes.OPCODE_IF_ICMPLT, 
				loopStartLabelOffset - loopGuardLabelOffset);
		instruction.setOffset(loopGuardLabelOffset);
		addInstruction(instructions, instruction, maxLocalStack);
		
		// load array on stack, so it can be used
		addInstruction(instructions,
				new ImmediateByteInstruction(Opcodes.OPCODE_ALOAD, false, arraySlot),
				maxLocalStack);
	}
	
	private List<PDSInstOld> getInstanceWrapperInstructions() 
		throws InvalidByteCodeException
	{
		PDSInfo pdsInfo = getInfo();
		List<PDSInstOld> instructions = new ArrayList<PDSInstOld>();
		int constructorIndex = indexOfMethodRef(classFile, "<init>", null);
		String constructorDesc = getDescriptor(classFile, constructorIndex);
		int paramCount = getParamCount(classFile, constructorIndex);
		// + 1 for heap index?
		// 3 for array initialization
		maxLocalStack = Math.max(3, Math.max(getParamCount(), paramCount) + 1);
		
		addInstruction(instructions, new ImmediateShortInstruction(Opcodes.OPCODE_NEW, 
				classFile.getThisClass()), maxLocalStack);
		
		addInstruction(instructions, new SimpleInstruction(Opcodes.OPCODE_DUP),
				maxLocalStack);
		
		addPushUndefinedParams(instructions, constructorDesc, paramCount, 
				maxLocalStack);

		addInstruction(instructions, new ImmediateShortInstruction(Opcodes.OPCODE_INVOKESPECIAL,
				constructorIndex), maxLocalStack);
		
		addPushUndefinedParams(instructions, method.getDescriptor(), 
				getParamCount(), maxLocalStack);
		
		addInstruction(instructions, new ImmediateShortInstruction(Opcodes.OPCODE_INVOKESPECIAL,
				indexOfMethodRef(classFile, method.getName(), method.getDescriptor())), maxLocalStack);

		addInstruction(instructions, new SimpleInstruction(Opcodes.OPCODE_RETURN),
				maxLocalStack);
		
		return instructions;
	}
	
	private void addPushUndefInstruction(List<PDSInstOld> instructions, int maxLocalStack)
		throws InvalidByteCodeException
	{
		addInstruction(instructions, new SimpleInstruction(Opcodes.OPCODE_ICONST_M1), 
				maxLocalStack);;
	}
	
	static int getParamCount(ClassFile classFile, int methodRefIndex)
		throws InvalidByteCodeException
	{
		ConstantMethodrefInfo info = 
			(ConstantMethodrefInfo)classFile.getConstantPoolEntry(methodRefIndex,
					ConstantMethodrefInfo.class);
		String desc = info.getNameAndTypeInfo().getDescriptor();
		return PDSMethod.countParam(desc);
	}
	
	static int indexOfMethodRef(ClassFile classFile, String name, String desc)
		throws InvalidByteCodeException
	{
		int index = findIndexOfMethodRef(classFile, name, desc);
		if (index != -1) {
			return index;
		}
		return createMethodRef(classFile, name, desc);
	}
	
	static int findIndexOfMethodRef(ClassFile classFile, String name, String desc)
		throws InvalidByteCodeException
	{
		CPInfo[] pool = classFile.getConstantPool();
		int candidate = -1;
		desc = desc != null ? desc : "()V";
		
		for (int i = 0; i < pool.length; i++) {
			CPInfo entry = pool[i];
			if (entry != null) {
				if (entry.getTag() == CPInfo.CONSTANT_METHODREF) {
					ConstantMethodrefInfo methodInfo = (ConstantMethodrefInfo)entry;
					if (methodInfo.getClassIndex() != classFile.getThisClass()) {
						// not a method ref of this class
						continue;
					}
					ConstantNameAndTypeInfo info = methodInfo.getNameAndTypeInfo();
					if (!info.getName().equals(name)) {
						// not a constructor
						continue;
					}
					// a constructor candidate
					candidate = i;
					if (info.getDescriptor().equals(desc)) {
						// the exact candidate
						return candidate;
					}
				}
			}
		}
		return candidate;
	}
	
	static int createMethodRef(ClassFile classFile, String name, String desc) 
		throws InvalidByteCodeException
	{
		int index = indexOfTypeAndName(classFile, name, desc);
		ConstantMethodrefInfo info = new ConstantMethodrefInfo();
		info.setClassFile(classFile);
		info.setNameAndTypeIndex(index);
		info.setClassIndex(classFile.getThisClass());
		return addConstantPoolEntry(classFile, info);
	}
	
	static int addConstantPoolEntry(ClassFile classFile, CPInfo info)
	{
		CPInfo[] pool = classFile.getConstantPool();
		CPInfo[] newPool = new CPInfo[pool.length + 1];
		System.arraycopy(pool, 0, newPool, 0, pool.length);
		newPool[newPool.length - 1] = info;
		classFile.setConstantPool(newPool);
		return newPool.length - 1;
	}
	
	/**
	 * 
	 * @param classFile
	 * @param name
	 * @param desc may be null
	 * @return
	 * @throws InvalidByteCodeException
	 */
	static int indexOfTypeAndName(ClassFile classFile, String name, String desc)
		throws InvalidByteCodeException
	{
		int index = findIndexOfMethodRef(classFile, name, desc);
		if (index != -1) {
			return index;
		}
		return createTypeAndName(classFile, name, desc);
	}
	
	static int findIndexOfTypeAndName(ClassFile classFile, String name, String desc)
		throws InvalidByteCodeException
	{
		CPInfo[] pool = classFile.getConstantPool();
		int candidate = -1;
		desc = desc != null ? desc : "()V";
		
		for (int i = 0; i < pool.length; i++) {
			CPInfo entry = pool[i];
			if (entry == null) {
				continue;
			}
			if (entry.getTag() == CPInfo.CONSTANT_NAME_AND_TYPE) {
				ConstantNameAndTypeInfo info = (ConstantNameAndTypeInfo)entry;
				if (!info.getName().equals(name)) {
					continue;
				}
				candidate = i;
				if (info.getDescriptor().equals(desc)) {
					// optimal constructor candidate;
					return candidate;
				}
			}
		}
		return candidate;
	}
	
	/**
	 * 
	 * @param classFile
	 * @param name
	 * @param desc may be null
	 * @return
	 * @throws InvalidByteCodeException
	 */
	static int createTypeAndName(ClassFile classFile, String name, String desc)
		throws InvalidByteCodeException
	{
		MethodInfo candidate = null;
		desc = desc != null ? desc : "()V";
		
		for (MethodInfo info : classFile.getMethods()) {
			if (!info.getName().equals(name)) {
				continue;
			}
			candidate = info;
			if (info.getDescriptor().equals(desc) ) {
				break;
			}
		}
		
		ConstantNameAndTypeInfo info = new ConstantNameAndTypeInfo();
		info.setClassFile(classFile);
		info.setNameIndex(candidate.getNameIndex());
		info.setDescriptorIndex(candidate.getDescriptorIndex());
		return addConstantPoolEntry(classFile, info);
	}

	private void appendLabel(StringBuilder out, int i)
	{
		out.append("\n").append(formattedName);
		if (i >= 0) {
			out.append(i);
		}
		out.append(": ");
	}
	
	private List<PDSInstOld> getStaticWrapperInstructions() 
		throws InvalidByteCodeException
	{
		PDSInfo pdsInfo = getInfo();
		List<PDSInstOld> instructions = new ArrayList<PDSInstOld>();
		int methodIndex = indexOfMethodRef(classFile, method.getName(),
				method.getDescriptor());
		// + 1 for heap index?
		// 3 for array initialization
		maxLocalStack = Math.max(3, getParamCount() + 1);
		
		addPushUndefinedParams(instructions, method.getDescriptor(), 
				getParamCount(), maxLocalStack);

		addInstruction(instructions, new ImmediateShortInstruction(Opcodes.OPCODE_INVOKESTATIC,
				methodIndex), maxLocalStack);
		
		addInstruction(instructions, new SimpleInstruction(Opcodes.OPCODE_RETURN), 
				maxLocalStack);
		
		return instructions;
	}
	
	/**
	 * Returns a new unique variable slot
	 * @return
	 */
	private int getNewVarSlot()
	{
		return varSlot++;
	}
	
	private int getNewLabelOffset()
	{
		return labelOffset++;
	}
	
	public String toRemoplaHead() throws InvalidByteCodeException 
	{
		StringBuilder out = new StringBuilder("module void ");
		return out.append(formattedName).append("()").toString();
	}
	
	public String getFormattedName() 
	{
		return formattedName;
	}
	
}
