/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;

public class PDS {
	
	static Logger logger = Logger.getLogger(PDS.class);
	
	private PDSInfo pdsInfo;
	private ArrayList<PDSClass> pdsClassList = new ArrayList<PDSClass>();

	/**
	 * Optional initial method which represents the entry point for the model
	 * checking analysis. This used to be the main method, but we want to
	 * allow other static and even non static methods as well.
	 */
	private PDSMethod initialMethod = null;
	
	private MethodWrapper initialMethodWrapper = null;
	
	/**
	 * Creates a PDS instance to generate Remopla code for the class file
	 * located at <code>classFileName</code>
	 * @param classFileName path name to the file can either be relative
	 * to the current working directory or absolute
	 * @throws Exception
	 */
	public PDS(String classFileName) throws Exception {
		this(new PDSInfo(new File(classFileName)));
	}
	
	public PDS(String classFileName, String methodName, String methodDescriptor)
		throws Exception
	{
		this(new PDSInfo(new File(classFileName), methodName, methodDescriptor));
	}
	
	/**
	 * Creates a PDS instance to generate Remopla code for the given
	 * class.
	 * @param clazz
	 * @throws Exception
	 */
	public PDS(Class clazz) throws Exception {
		this(new PDSInfo(clazz));
	}
	
	public PDS(Class clazz, String methodName, String methodDescriptor) 
		throws Exception
	{
		this(new PDSInfo(clazz, methodName, methodDescriptor));
	}

	public PDS(String className, String[] pathEntries) throws Exception 
	{
		this(new PDSInfo(className, pathEntries));
	}
	
	public PDS(String className, String[] pathEntries, String methodName,
			String methodDescriptor) throws Exception 
	{
		this(new PDSInfo(className, pathEntries, methodName, methodDescriptor));
	}
	
	public PDS(PDSInfo info) throws Exception {
		pdsInfo = info;
		logger.info("Entering PDS constructor.");
		initPDSClassList();
		setupInitialMethod();
		logger.info("Leaving PDS constructor.");
	}
	
	public void translate() throws InvalidByteCodeException {
		
		for (PDSClass pdsClass : pdsClassList) {
			pdsClass.translate();
		}
	}
	
	/**
	 * Returns the configuration object responsible for this pds instance.
	 * @return
	 */
	public PDSInfo getInfo()
	{
		return pdsInfo;
	}
	
	private void setupInitialMethod() throws InvalidByteCodeException, IOException
	{
		MethodInfo method = pdsInfo.getInitialMethod();
		if (method == null) {
			return;
		}
		
		initialMethod = pdsClassList.get(0).getMethod(method.getName(),
				method.getDescriptor());
		
		if (!initialMethod.isMainMethod()) {
			if (!initialMethod.isStatic()|| initialMethod.getParamCount() > 0) {
				initialMethodWrapper = new MethodWrapper(initialMethod.getMethodInfo(), 
						pdsInfo);
			}
		}
	}
	
	/**
	 * Returns the method that has been set to be the entry point for the analysis or
	 * the main method of the class that is to be analysed.
	 * 
	 * @return <code>null</code> if no method has been set and there is no main 
	 * method in the first class
	 * @throws InvalidByteCodeException 
	 */
	public PDSMethod getInitialMethod()
	{
		return initialMethod;
	}
	
	/**
	 * Returns null if there is none.
	 */
	public MethodWrapper getInitialMethodWrapper()
	{
		return initialMethodWrapper;
	}
	
	private PDSMethod getInternalInitialMethod()
	{
		return initialMethodWrapper != null ? initialMethodWrapper : getInitialMethod();
	}
	
	public String toRemopla() throws InvalidByteCodeException, IOException {
		
		PDSClass pdsClass;
		StringBuffer out = new StringBuffer();
		PDSMultiArrayInitializerMethod multiArrayMethod = 
			pdsInfo.includeMultiArrayInitializerMethod() ? 
					new PDSMultiArrayInitializerMethod(pdsInfo) : null;
		
		int heapSize = pdsInfo.getStoredHeapSize();
		int heapBit = PDSDefault.getMinimumBitsForHeapSize(heapSize);
		int intBit = pdsInfo.getBits();
		
		out.append("define DEFAULT_INT_BITS ");
		out.append(String.valueOf(intBit));
		out.append("\n\n");
		
		out.append("int ");
		out.append(pdsInfo.heapIndex(String.valueOf(pdsInfo.getHeapSize())));
		if (pdsInfo.getHeapOption() == PDSInfo.HeapOption.TWODIMS) {
			out.append("(1)");
		}
		out.append(";\n");
		
		out.append("int ");
		out.append(pdsInfo.getHeapPtrName());
		out.append("(");
		out.append(String.valueOf(heapBit));
		out.append(");\n\n");
		
		// collect declarations of static fields and declarations of
		// class methods
		
		if (initialMethodWrapper != null) {
			out.append(initialMethodWrapper.toRemoplaHead()).append(";\n");
		}
		if (multiArrayMethod != null 
				&& pdsInfo.getHeapOption() == PDSInfo.HeapOption.SIMPLE) {
			out.append(multiArrayMethod.toRemoplaHead()).append(";\n");
		}
		if (pdsInfo.isIncludeStringEqualsMethod()) {
			out.append(PDSString.remoplaHeadForEquals()).append(";\n");
		}
		
		for (int i = 0; i < pdsClassList.size(); i++) {
			
			pdsClass = (PDSClass) pdsClassList.get(i);
			out.append(pdsClass.toRemoplaGlobal());
			out.append("\n");
		}
		
		out.append("init");
		
		if (PDSInfo.Stefan) {
			addStefansOutput(out);
		} 
		// We'll always use Stefan's option.
		/*else {

			for (int i = pdsClassList.size() - 1; i >= 0; i--) {
				
				pdsClass = (PDSClass) pdsClassList.get(i);
				pdsMethod = pdsClass.getMethodClinit();
				if (pdsMethod != null) {
					pdsMethod.addInitPtr();	//Dirty pointer initialization
					out.append(" ");
					out.append(pdsMethod.getFormattedName());
				}
			}
			pdsMethod = getInternalInitialMethod();
			if (pdsMethod != null) {
				out.append(" ");
				out.append(pdsMethod.getFormattedName());
			}
			out.append(";\n\n");
		}*/
		
		// collect all method definitions
		
		if (initialMethodWrapper != null) {
			out.append(initialMethodWrapper.toRemopla()).append("\n");
		}
		if (multiArrayMethod != null 
				&& pdsInfo.getHeapOption() == PDSInfo.HeapOption.SIMPLE) {
			out.append(multiArrayMethod.toRemopla()).append("\n");
		}
		if (pdsInfo.isIncludeStringEqualsMethod()) {
			out.append(PDSString.remoplaForEquals(pdsInfo)).append("\n");
		}
		
		for (int i = 0; i < pdsClassList.size(); i++) {
			pdsClass = (PDSClass) pdsClassList.get(i);
			out.append(pdsClass.toRemoplaModule());
			out.append("\n");
		}

		// add assert error state
		// TODO maybe only add this if it might be reachable, i.e. if there are 
		// assert statements in the code
		out.append(selfLoopStmt(pdsInfo.getLabelAssertError()));
		
		out.append(selfLoopStmt(pdsInfo.getLabelGeneralException()));
		
		if (pdsInfo.checkForHeapOverflow()) {
			out.append(selfLoopStmt(pdsInfo.getLabelHeapOverflow()));
		}
		
		if (pdsInfo.checkForIndexOutOfBounds()) {
			out.append(selfLoopStmt(pdsInfo.getLabelIndexOutOfBounds()));
		}
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			out.append(selfLoopStmt(pdsInfo.getLabelNPE()));
		}
		
		if (pdsInfo.checkForStringBuilderOverflow()) {
			out.append(selfLoopStmt(pdsInfo.getLabelStringBuilderOverflow()));
		}
		
		return out.toString();
	}
	
	public void addStefansOutput(StringBuffer out) {

		PDSClass tmpClass;
		PDSMethod tmpMethod;
		String initMethodName = null;
		PDSClass pdsClass;
		PDSMethod pdsMethod;
		
		out.append(" ");
		out.append(PDSDefault.JMOPED_INIT);
		for (int i = pdsClassList.size() - 1; i >= 0; i--) {
			
			pdsClass = (PDSClass) pdsClassList.get(i);
			pdsMethod = pdsClass.getMethodClinit();
			if (pdsMethod != null) {
				
				if (initMethodName == null) {
					initMethodName = pdsMethod.getFormattedName();
				}
				tmpMethod = null;
				for (int j = i - 1; j >= 0; j--) {
					tmpClass = (PDSClass) pdsClassList.get(j);
					tmpMethod = tmpClass.getMethodClinit();
					if (tmpMethod != null) {
						break;
					}
				}
				
				if (tmpMethod != null) {
					pdsMethod.replaceLastInst("goto " + tmpMethod.getFormattedName());
				} else {
					tmpMethod = getInternalInitialMethod();
					if (tmpMethod != null) {
						pdsMethod.replaceLastInst("goto " + tmpMethod.getFormattedName());
					} else {
						logger.error("Main method not found.");
					}
				}
			}
		}
		if (initMethodName == null) {
			pdsMethod = getInternalInitialMethod();
			if (pdsMethod != null) {
				initMethodName = pdsMethod.getFormattedName();
			} else {
				logger.error("Main method not found.");
			}
		}
		out.append(";\n\n");
		out.append(initModule(initMethodName));
	}
	
	// TODO move this to unit tests
	public static void main(String[] args) {
		
		PDS pds;
		
		try {
			PDSInfo.simplified = true;
			pds = new PDS("ex/LinkedList.class");
			logger.info("Included Methods:\n" + pds.includedTableInfo());
			logger.info("Field Indices:\n" + pds.fieldIndexTableInfo());
			System.out.print(pds.toRemopla());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/*******************************************************************/
	/**                       PDSInfo METHODS                         **/
	/*******************************************************************/
	
	public String fieldIndexTableInfo() {
		
		return pdsInfo.fieldIndexTableInfo();
	}
	
	public String getLabelAssertError() {
		
		return pdsInfo.getLabelAssertError();
	}
	
	public String getLabelHeapOverflow() {
		
		return pdsInfo.getLabelHeapOverflow();
	}

	public String getLabelNPE() {
	
	return pdsInfo.getLabelNPE();
	}
	
	public String getMopedPath() {
		
		return pdsInfo.getMopedPath();
	}
	
	public String includedTableInfo() throws InvalidByteCodeException {
		
		return pdsInfo.includedTableInfo();
	}
	
	public void putAllLineBitTable(Hashtable<Integer, Integer> lbTable) {
		
		pdsInfo.putAllLineBitTable(lbTable);
	}
	
	public void setBits(int bits) {
		
		pdsInfo.setBits(bits);
	}
	
	public void setHeapOption(PDSInfo.HeapOption heapOption) {
		
		pdsInfo.setHeapOption(heapOption);
	}
	
	public void setHeapSize(int heapSize) {
		
		pdsInfo.setHeapSize(heapSize);
	}
	
	/*******************************************************************/
	/**                       PRIVATE METHODS                         **/
	/*******************************************************************/
	
	private String selfLoopStmt(String label) {
		
		StringBuffer out = new StringBuffer();
		
		out.append(label);
		out.append(": goto ");
		out.append(label);
		out.append(";\n");
		
		return out.toString();
	}
	
	/**
	 * Create a special module, used to initialize heap and its pointer.
	 * The module name is defined in PDSDefault.JMOPED_INIT.
	 * The method is called when Stefan option is on.
	 * After the initialization, the instruction "goto gotoName" is added.
	 * 
	 * @param gotoName
	 * @return
	 */
	private String initModule(String gotoName) {
		
		StringBuffer out = new StringBuffer();
		
		out.append("module void ");
		out.append(PDSDefault.JMOPED_INIT);
		out.append("()\n");
		out.append("{\n");
		out.append(pdsInfo.getHeapPtrName());
//		out.append("=1;\n");
		out.append("=1, ");
		out.append("A i (0,");
		out.append(String.valueOf(pdsInfo.getHeapSize() - 1));
		out.append(") ");
		out.append(pdsInfo.heapIndex("i"));
		out.append("=0;\n");
		out.append("goto ");
		out.append(gotoName);
		out.append(";\n");
		out.append("}\n\n");
		
		return out.toString();
	}
	

	/**
	 * Returns an unmodifiable list of all PDSClass objects that have been
	 * created.
	 *
	 * @return 
	 */
	public List<PDSClass> getPDSClassList()
	{
		return Collections.unmodifiableList(pdsClassList);
	}
	
	/**
	 * @return null in case it's not found
	 */
	public PDSMethod getMethodByFormattedName(String methodName)
	{
		
		if (initialMethodWrapper != null 
				&& initialMethodWrapper.getFormattedName().equals(methodName)) {
			return initialMethodWrapper;
		}
		
		for (PDSClass clazz : pdsClassList) {
			if (methodName.startsWith(clazz.getFormattedClassName())) {
				for (PDSMethod method : clazz.getMethodList()) {
					if (method.getFormattedName().equals(methodName)) {
						return method;
					}
				}
			}
		}
		return null;
	}
	
	private void initPDSClassList() throws InvalidByteCodeException, IOException
	{
		logger.debug("Entering initPDSClassList");
		
		List<ClassFile> classFileList = pdsInfo.getClassFileList();
		for (ClassFile classFile : classFileList) {
			pdsClassList.add(new PDSClass(classFile, pdsInfo));
		}
		
		pdsInfo.buildFieldTable(pdsClassList);
		pdsInfo.buildClassTable(pdsClassList);
		
		for (PDSClass pdsClass : pdsClassList) {
			pdsClass.initPDSMethodList(pdsInfo);
		}
			
		logger.debug("Leaving initPDSClassList");
	}
	
}
