/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import org.gjt.jclasslib.bytecode.*;
import org.gjt.jclasslib.structures.*;
import org.gjt.jclasslib.structures.constants.*;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.List;

import javax.swing.tree.DefaultMutableTreeNode;

public class PDSInstOld {
	
	static Logger logger = Logger.getLogger(PDSInstOld.class);
	
	private ClassFile classFile;
	private String methodName;
	private AbstractInstruction absInst = null;
	private int maxStack;
	private PDSInfo pdsInfo;
	private int type;
	private ArrayList inst;
	/**
	 * Marks if this instruction was replaced from the outside. If this is the case, the
	 * Remopla output will not be rebuilt in {@link #toRemopla()} and the actual 
	 * underlying bytecode instruction will be ignored.
	 */
	private boolean instReplaced = false;

	static final int NORMAL_INST = 0;
	static final int IF_INST = 1;
	static final int DOUBLE_INST = 2;
	
	public PDSInstOld(ClassFile classFile, String methodName, 
			AbstractInstruction absInst, int maxStack, PDSInfo pdsInfo)
		throws InvalidByteCodeException 
	{
		this.classFile = classFile;
		this.methodName = methodName;
		this.absInst = absInst;
		this.maxStack = maxStack;
		this.pdsInfo = pdsInfo;
		type = NORMAL_INST;

		logger.debug(toRemopla());
	}
	
	public PDSInstOld(ClassFile classFile, String methodName, PDSInfo pdsInfo) 
	{
		this.classFile = classFile;
		this.methodName = methodName;
		this.pdsInfo = pdsInfo;
		type = NORMAL_INST;
		inst = new ArrayList();
	}
	
	AbstractInstruction getInstruction()
	{
		return absInst;
	}
	
	/** Dirty hack to init heap pointer. To be fixed. */
	public void addInitPtr() {
		
		addInst(pdsInfo.getHeapPtrName() + "=1");
	}
	
	public void addInst(String stmt) {
		
		inst.add(stmt);
//		logger.debug(toRemopla());
	}
	
	public void replaceInst(String st) {
		
		inst = new ArrayList();
		inst.add(st);
		instReplaced = true;
	}
	
	/**
	 * Returns the label for this instruction.
	 * <p>
	 * If the pc offset in the method is large than 0, it is appended, otherwise just
	 * the formatted method name is returned.
	 */
	private String getLabel() 
	{
		StringBuilder builder = new StringBuilder(methodName);
		if ((absInst != null) && absInst.getOffset() > 0) {
			builder.append(absInst.getOffset());
		}
		return builder.toString();
	}
	
	public String toRemopla() throws InvalidByteCodeException {
		
		if (absInst != null && !instReplaced) {
			inst = null;
			inst = new ArrayList();
			initInst();
		}
		
		StringBuilder out = new StringBuilder();

		out.append(getLabel());
		out.append(":\t");
		switch (type) {
			
			case NORMAL_INST:
				out.append(commaList(inst, inst.size()));
				out.append(";");
				break;
				
			case IF_INST: {
				out.append("if\n");
				ArrayList[] ifthen;
				for (int i = 0; i < inst.size() ; i++) {
					if (!(inst.get(i) instanceof ArrayList[])) {
						System.out.println(absInst.getOpcodeVerbose() + " " + absInst);
						System.out.println(inst.get(i));
					}
					ifthen = (ArrayList[]) inst.get(i);
					out.append("\t:: ");
					out.append((String) ifthen[0].get(0));
					out.append(" -> ");
					out.append(commaList(ifthen[1], ifthen[1].size()));
					out.append(";\n");
				}
				out.append("fi");
				break;
			}
				
			case DOUBLE_INST:
				out.append(commaList(inst, inst.size() - 1));
				out.append("; ");
				out.append((String) inst.get(inst.size() - 1));
				out.append(";");
				break;
		}
		return out.toString();
	}
	
	/**
	 * Concatenates elements of the list to a comma separated list and 
	 * returns it as a string.
	 *
	 * @throws IndexOutOfBoundsException if <code>to > list.size() </code>
	 */
	public static String commaList(List list, int to) {

		if (list.size() == 0) {
			logger.error("Error: size of instruction list (inst) = 0");
		}

		if (list.size() < to) {
			throw new IndexOutOfBoundsException(
					"to is greater than size of list");
		}
		StringBuilder out = new StringBuilder();
		Iterator iter = list.iterator();
		Object next = iter.next();
		
		appendToBuilder(out, next);

		for (int i = 1; i < to; i++) {
			out.append(", ");
			appendToBuilder(out, iter.next());
		}
		
		return out.toString();
	}
	
	private static void appendToBuilder(StringBuilder out, Object next)
	{
		if (next instanceof List[]) {
			List[] innerLists = (List[])next;
			for (List l : innerLists) {
				for (Object o : l) {
					out.append(o);
				}
			}
		}
		else {
			out.append(next);
		}		
	}
	
	static String[] replaceClassName(String[] names, String name) {
		
		String[] newNames = new String[names.length];
		newNames[0] = name;
		newNames[1] = names[1];
		if (names.length == 3) {
			newNames[2] = names[2];
		}
		
		return newNames;
	}
	
	private void initInst() throws InvalidByteCodeException {
		
		switch (absInst.getOpcode()) {
			
			case Opcodes.OPCODE_AALOAD: 
				ialoadInst();
				break;
			case Opcodes.OPCODE_AASTORE:
				iastoreInst();
				break;
			case Opcodes.OPCODE_ACONST_NULL: 
				pushInst(0);
				break;
			case Opcodes.OPCODE_ALOAD: 
				pushInst(pdsInfo.varIndex(immediateByte()));
				break;
			case Opcodes.OPCODE_ALOAD_0: 
			case Opcodes.OPCODE_ALOAD_1: 
			case Opcodes.OPCODE_ALOAD_2: 
			case Opcodes.OPCODE_ALOAD_3: 
				pushInst(pdsInfo.varIndex(
						absInst.getOpcode() - Opcodes.OPCODE_ALOAD_0));
				break;
			case Opcodes.OPCODE_ANEWARRAY:
				newarrayInst();
				break;
			case Opcodes.OPCODE_ARETURN: 
				inst.add("return " + pdsInfo.stackIndex(0));
				break;
			case Opcodes.OPCODE_ARRAYLENGTH: 
				arraylengthInst();
				break;
			case Opcodes.OPCODE_ASTORE:
				popInst(pdsInfo.varIndex(immediateByte()));
				break;
			case Opcodes.OPCODE_ASTORE_0: 
			case Opcodes.OPCODE_ASTORE_1: 
			case Opcodes.OPCODE_ASTORE_2: 
			case Opcodes.OPCODE_ASTORE_3:
				popInst(pdsInfo.varIndex(
						absInst.getOpcode() - Opcodes.OPCODE_ASTORE_0));
				break;
			case Opcodes.OPCODE_ATHROW: 
				//TODO
				inst.add("skip");
				break;
			case Opcodes.OPCODE_BALOAD: 
				ialoadInst();
				break;
			case Opcodes.OPCODE_BASTORE:
				iastoreInst();
				break;
			case Opcodes.OPCODE_BIPUSH: 
				pushInst(immediateByte());
				break;
			case Opcodes.OPCODE_CALOAD: 
				ialoadInst();
				break;
			case Opcodes.OPCODE_CASTORE:
				iastoreInst();
				break;
			case Opcodes.OPCODE_CHECKCAST: 
				inst.add("skip");
				break;
			case Opcodes.OPCODE_DADD: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DALOAD: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DASTORE:
				popPushInst(3, 0);
				break;
			case Opcodes.OPCODE_DCMPG: 
			case Opcodes.OPCODE_DCMPL: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DCONST_0: 
			case Opcodes.OPCODE_DCONST_1: 
				pushInst(absInst.getOpcode() - Opcodes.OPCODE_DCONST_0);
				break;
			case Opcodes.OPCODE_DDIV: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DLOAD: 
			case Opcodes.OPCODE_DLOAD_0: 
			case Opcodes.OPCODE_DLOAD_1: 
			case Opcodes.OPCODE_DLOAD_2: 
			case Opcodes.OPCODE_DLOAD_3: 
				pushInst("undef");
				break;
			case Opcodes.OPCODE_DMUL: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DNEG: 
				inst.add(pdsInfo.stackIndex(0) + "=undef");
				break;
			case Opcodes.OPCODE_DREM: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DRETURN: 
				inst.add("return " + pdsInfo.stackIndex(0));
				break;
			case Opcodes.OPCODE_DSTORE: 
			case Opcodes.OPCODE_DSTORE_0: 
			case Opcodes.OPCODE_DSTORE_1: 
			case Opcodes.OPCODE_DSTORE_2: 
			case Opcodes.OPCODE_DSTORE_3: 
				popInst(1, 1);
				break;
			case Opcodes.OPCODE_DSUB: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_DUP: 
				pushInst(pdsInfo.stackIndex(0));
				break;
			case Opcodes.OPCODE_DUP_X1: 
				inst.add(stackEqual(0, 0));
				inst.add(stackEqual(1, 1));
				inst.add(stackEqual(2, 0));
				pushInst(3, 1);
				break;
			case Opcodes.OPCODE_DUP_X2: 
				inst.add(stackEqual(0, 0));
				inst.add(stackEqual(1, 1));
				inst.add(stackEqual(2, 2));
				inst.add(stackEqual(3, 0));
				pushInst(4, 1);
				break;
			case Opcodes.OPCODE_DUP2: 
				inst.add(stackEqual(0, 0));
				inst.add(stackEqual(1, 1));
				inst.add(stackEqual(2, 0));
				inst.add(stackEqual(3, 1));
				pushInst(4, 1);
				break;
			case Opcodes.OPCODE_FADD: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_FALOAD: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_FASTORE:
				popPushInst(3, 0);
				break;
			case Opcodes.OPCODE_FCMPG: 
			case Opcodes.OPCODE_FCMPL: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_FCONST_0: 
			case Opcodes.OPCODE_FCONST_1: 
			case Opcodes.OPCODE_FCONST_2: 
				pushInst(absInst.getOpcode() - Opcodes.OPCODE_FCONST_0);
				break;
			case Opcodes.OPCODE_FDIV: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_FLOAD: 
			case Opcodes.OPCODE_FLOAD_0: 
			case Opcodes.OPCODE_FLOAD_1: 
			case Opcodes.OPCODE_FLOAD_2: 
			case Opcodes.OPCODE_FLOAD_3: 
				pushInst("undef");
				break;
			case Opcodes.OPCODE_FMUL: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_FNEG: 
				inst.add(pdsInfo.stackIndex(0) + "=undef");
				break;
			case Opcodes.OPCODE_FREM: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_FRETURN: 
				inst.add("return " + pdsInfo.stackIndex(0));
				break;
			case Opcodes.OPCODE_FSTORE: 
			case Opcodes.OPCODE_FSTORE_0: 
			case Opcodes.OPCODE_FSTORE_1: 
			case Opcodes.OPCODE_FSTORE_2: 
			case Opcodes.OPCODE_FSTORE_3: 
				popInst(1, 1);
				break;
			case Opcodes.OPCODE_FSUB: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_GETFIELD: 
				getfieldInst();
				break;
			case Opcodes.OPCODE_GETSTATIC: 
				getstaticInst();
				break;
			case Opcodes.OPCODE_GOTO: 
				inst.add(gotoStmt());
				break;
			case Opcodes.OPCODE_IADD: 
				arithInst("+");
				break;
			case Opcodes.OPCODE_IALOAD: 
				ialoadInst();
				break;
			case Opcodes.OPCODE_IAND: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_IASTORE:
				iastoreInst();
				break;
			case Opcodes.OPCODE_ICONST_0: 
			case Opcodes.OPCODE_ICONST_1: 
			case Opcodes.OPCODE_ICONST_2: 
			case Opcodes.OPCODE_ICONST_3: 
			case Opcodes.OPCODE_ICONST_4: 
			case Opcodes.OPCODE_ICONST_5: 
				pushInst(absInst.getOpcode() - Opcodes.OPCODE_ICONST_0);
				break;
			case Opcodes.OPCODE_ICONST_M1: 
				pushInst("undef");
				break;
			case Opcodes.OPCODE_IDIV: 
				arithInst("/");
				break;
			case Opcodes.OPCODE_IF_ACMPEQ:
				ifIcmpInst("==");
				break;
			case Opcodes.OPCODE_IF_ACMPNE:
				ifIcmpInst("!=");
				break;
			case Opcodes.OPCODE_IF_ICMPEQ:
				ifIcmpInst("==");
				break;
			case Opcodes.OPCODE_IF_ICMPGE:
				ifIcmpInst(">=");
				break;
			case Opcodes.OPCODE_IF_ICMPGT:
				ifIcmpInst(">");
				break;
			case Opcodes.OPCODE_IF_ICMPLE:
				ifIcmpInst("<=");
				break;
			case Opcodes.OPCODE_IF_ICMPLT:
				ifIcmpInst("<");
				break;
			case Opcodes.OPCODE_IF_ICMPNE:
				ifIcmpInst("!=");
				break;
			case Opcodes.OPCODE_IFEQ: 
				ifInst("==");
				break;
			case Opcodes.OPCODE_IFGE: 
				ifInst(">=");
				break;
			case Opcodes.OPCODE_IFGT: 
				ifInst(">");
				break;
			case Opcodes.OPCODE_IFLE: 
				ifInst("<=");
				break;
			case Opcodes.OPCODE_IFLT: 
				ifInst("<");
				break;
			case Opcodes.OPCODE_IFNE: 
				ifInst("!=");
				break;
			case Opcodes.OPCODE_IFNONNULL: 
				ifInst("!=");
				break;
			case Opcodes.OPCODE_IFNULL: 
				ifInst("==");
				break;
			case Opcodes.OPCODE_IINC: 
				iincInst();
				break;
			case Opcodes.OPCODE_ILOAD: 
				pushInst(pdsInfo.varIndex(immediateByte()));
				break;
			case Opcodes.OPCODE_ILOAD_0: 
			case Opcodes.OPCODE_ILOAD_1: 
			case Opcodes.OPCODE_ILOAD_2: 
			case Opcodes.OPCODE_ILOAD_3: 
				pushInst(pdsInfo.varIndex(
						absInst.getOpcode() - Opcodes.OPCODE_ILOAD_0));
				break;
			case Opcodes.OPCODE_IMUL: 
				arithInst("*");
				break;
			case Opcodes.OPCODE_INEG: 
				inst.add(pdsInfo.stackIndex(0) + "=undef");
				break;
			case Opcodes.OPCODE_INSTANCEOF: 
				//TODO
				inst.add("skip");
				break;
			case Opcodes.OPCODE_INVOKEINTERFACE:
				invokeinterfaceInst();
				break;
			case Opcodes.OPCODE_INVOKESPECIAL: 
				invokespecialInst();
				break;
			case Opcodes.OPCODE_INVOKESTATIC:
				invokestaticInst();
				break;
			case Opcodes.OPCODE_INVOKEVIRTUAL: 
				invokevirtualInst();
				break;
			case Opcodes.OPCODE_IOR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_IREM: 
				iremInst();
				break;
			case Opcodes.OPCODE_IRETURN: 
				inst.add("return " + pdsInfo.stackIndex(0));
				break;
			case Opcodes.OPCODE_ISHL: 
				arithInst("<<");
				break;
			case Opcodes.OPCODE_ISHR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_ISTORE:
				popInst(pdsInfo.varIndex(immediateByte()));
				break;
			case Opcodes.OPCODE_ISTORE_0: 
			case Opcodes.OPCODE_ISTORE_1: 
			case Opcodes.OPCODE_ISTORE_2: 
			case Opcodes.OPCODE_ISTORE_3: 
				popInst(pdsInfo.varIndex(
						absInst.getOpcode() - Opcodes.OPCODE_ISTORE_0));
				break;
			case Opcodes.OPCODE_ISUB: 
				arithInst("-");
				break;
			case Opcodes.OPCODE_IUSHR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_IXOR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_JSR: 
				type = DOUBLE_INST;
				pushInst(absInst.getOffset() + 2);
				inst.add(gotoStmt());
				break;
			case Opcodes.OPCODE_JSR_W: 
				type = DOUBLE_INST;
				pushInst(absInst.getOffset() + 4);
				inst.add(gotoStmt());
				break;
			case Opcodes.OPCODE_LADD: 
				arithInst("+");
				break;
			case Opcodes.OPCODE_LALOAD: 
				ialoadInst();
				break;
			case Opcodes.OPCODE_LAND: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_LASTORE:
				iastoreInst();
				break;
			case Opcodes.OPCODE_LCMP:
				lcmpInst();
				break;
			case Opcodes.OPCODE_LCONST_0: 
			case Opcodes.OPCODE_LCONST_1: 
				pushInst(absInst.getOpcode() - Opcodes.OPCODE_LCONST_0);
				break;
			case Opcodes.OPCODE_LDC: 
				ldcInst();
				break;
			case Opcodes.OPCODE_LDC_W: 
			case Opcodes.OPCODE_LDC2_W: 
				pushInst("undef");
				break;
			case Opcodes.OPCODE_LDIV: 
				arithInst("/");
				break;
			case Opcodes.OPCODE_LLOAD: 
				pushInst(pdsInfo.varIndex(immediateByte()));
				break;
			case Opcodes.OPCODE_LLOAD_0: 
			case Opcodes.OPCODE_LLOAD_1: 
			case Opcodes.OPCODE_LLOAD_2: 
			case Opcodes.OPCODE_LLOAD_3: 
				pushInst(pdsInfo.varIndex(
						absInst.getOpcode() - Opcodes.OPCODE_LLOAD_0));
				break;
			case Opcodes.OPCODE_LMUL: 
				arithInst("*");
				break;
			case Opcodes.OPCODE_LNEG: 
				inst.add(pdsInfo.stackIndex(0) + "=undef");
				break;
			case Opcodes.OPCODE_LOOKUPSWITCH:
				lookupswitchInst();
				break;
			case Opcodes.OPCODE_LOR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_LREM: 
				iremInst();
				break;
			case Opcodes.OPCODE_LRETURN: 
				inst.add("return " + pdsInfo.stackIndex(0));
				break;
			case Opcodes.OPCODE_LSHL: 
				arithInst("<<");
				break;
			case Opcodes.OPCODE_LSHR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_LSTORE:
				popInst(pdsInfo.varIndex(immediateByte()));
				break;
			case Opcodes.OPCODE_LSTORE_0: 
			case Opcodes.OPCODE_LSTORE_1: 
			case Opcodes.OPCODE_LSTORE_2: 
			case Opcodes.OPCODE_LSTORE_3: 
				popInst(pdsInfo.varIndex(
						absInst.getOpcode() - Opcodes.OPCODE_LSTORE_0));
				break;
			case Opcodes.OPCODE_LSUB: 
				arithInst("-");
				break;
			case Opcodes.OPCODE_LUSHR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_LXOR: 
				popPushInst(2, 1);
				break;
			case Opcodes.OPCODE_MULTIANEWARRAY: 
				multianewarrayInst();
				break;
			case Opcodes.OPCODE_NEW: 
				newInst();
				break;
			case Opcodes.OPCODE_NEWARRAY: 
				newarrayInst();
				break;
			case Opcodes.OPCODE_NOP: 
				inst.add("skip");
				break;
			case Opcodes.OPCODE_POP:
				popInst(1, 1);
				break;
			case Opcodes.OPCODE_POP2:
				popInst(2, 2);
				break;
			case Opcodes.OPCODE_PUTFIELD:
				putfieldInst();
				break;
			case Opcodes.OPCODE_PUTSTATIC: 
				putstaticInst();
				break;
			case Opcodes.OPCODE_RET: 
				retInst();
				break;
			case Opcodes.OPCODE_RETURN:
				inst.add("return");
				break;
			case Opcodes.OPCODE_SALOAD: 
				ialoadInst();
				break;
			case Opcodes.OPCODE_SASTORE:
				iastoreInst();
				break;
			case Opcodes.OPCODE_SIPUSH: 
				pushInst(((ImmediateShortInstruction) absInst).getImmediateShort());
				break;
			case Opcodes.OPCODE_SWAP: 
				inst.add(stackEqual(0, 1));
				inst.add(stackEqual(1, 0));
				break;
			case Opcodes.OPCODE_TABLESWITCH: 
				tableswitchInst();
				break;
		}
		
		if (inst.size() == 0) {
			inst.add("skip");
		}
	}
	
	/*******************************************************************/
	/**              FUNCTIONS FOR COMMON INSTRUCTIONS                **/
	/*******************************************************************/
	
	private void arithInst(String op) {
		
		inst.add(stackArith(op));
		popInst(2, 1);
	}
	
	private void arraylengthInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			type = IF_INST;
			inst.add(ifStmt(s0 + "==0", gotoStmt(pdsInfo.getLabelNPE())));
			inst.add(ifStmt("else", s0 + "=" + pdsInfo.heapIndex(s0)));
		} else {
			inst.add(s0 + "=" + pdsInfo.heapIndex(s0));
		}
	}
	
	private void getfieldInst() throws InvalidByteCodeException {
		
		String[] fieldName = getReferencedName(classFile, absInst);
		
		if (pdsInfo.isIgnoredField(fieldName)) {
			if (pdsInfo.checkForNullPointerExceptions()) {
				popPushGTZStmt(1,1);
			} else {
				popPushInst(1, 1);
			}
			return;
		}
		
		String s0 = pdsInfo.stackIndex(0);
		DefaultMutableTreeNode thisNode = pdsInfo.getClassNode(fieldName[0]);
		Enumeration e = thisNode.breadthFirstEnumeration();
		DefaultMutableTreeNode subNode;
		String[] subName;
		String formattedSubName;
		String objType = pdsInfo.heapIndex(s0);
		int fieldIndex;
		String h;
		type = IF_INST;
		
		while (e.hasMoreElements()) {
			subNode = (DefaultMutableTreeNode) e.nextElement();
			subName = replaceClassName(fieldName, 
					(String) subNode.getUserObject());
			formattedSubName = PDSField.formatFieldName(subName);
			fieldIndex = pdsInfo.indexOfField(formattedSubName) + 1;
			if (fieldIndex > 0) {
				h = pdsInfo.heapIndex(s0 + "+" + fieldIndex);
				inst.add(ifStmt(
						objType + "==" + pdsInfo.getClassIndex(subName[0]),
						popPushStmt(h, 1, 1)));
			}
		}
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			inst.add(ifStmt(s0 + "==0", gotoStmt(pdsInfo.getLabelNPE())));
		}
	}
	
	private void getstaticInst() throws InvalidByteCodeException {
		
		String[] fieldName = getReferencedName(classFile, absInst);
		if (pdsInfo.isIgnoredField(fieldName)) {
			pushInst("undef");
		} else {
			pushInst(PDSField.formatFieldName(fieldName));
		}
	}
	
	private void ialoadInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		String h = pdsInfo.heapIndex(s1 + "+" + s0 + "+1");
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			type = IF_INST;
			inst.add(ifStmt(s1 + "==0", gotoStmt(pdsInfo.getLabelNPE())));
			inst.add(ifStmt("else", popPushStmt(h, 2, 1)));
		} else {
			popPushInst(h, 2, 1);
		}
	}
	
	private void iastoreInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		String s2 = pdsInfo.stackIndex(2);
		String h = pdsInfo.heapIndex(s2 + "+" + s1 + "+1");
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			type = IF_INST;
			inst.add(ifStmt(s2 + "==0", gotoStmt(pdsInfo.getLabelNPE())));
			inst.add(ifStmt("else", popPushStmt(h + "=" + s0, 3, 0)));
		} else {
			popPushInst(h + "=" + s0, 3, 0);
		}
	}
	
	private void ifIcmpInst(String op) {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		
		type = IF_INST;
		inst.add(ifStmt(s1 + op + s0, gotoStmt(), 2, 2));
		inst.add(ifStmt("else", null, 2, 2));
	}
	
	private void ifInst(String op) {
		
		type = IF_INST;
		inst.add(ifStmt(pdsInfo.stackIndex(0) + op + 0, gotoStmt(), 1, 1));
		inst.add(ifStmt("else", null, 1, 1));
	}
	
	private void iincInst() {
		
		IncrementInstruction incInst = (IncrementInstruction) absInst;
		int index = incInst.getImmediateByte();
		int cnst = incInst.getIncrementConst();
		
		if (cnst <= 127) {
			inst.add(pdsInfo.varIndex(index) + "=" + pdsInfo.varIndex(index) 
				+ "+" + String.valueOf(cnst));
		} else {
			inst.add(pdsInfo.varIndex(index) + "=" + pdsInfo.varIndex(index) 
				+ "-" + String.valueOf(256 - cnst));
		}
	}
	
	private void invokeinterfaceInst() throws InvalidByteCodeException {
		
		String[] calledName = calleeInterfaceName();
		int paramCount = PDSMethod.countParam(calledName[2]) + 1;
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		Collection<String> impList = pdsInfo.getImplementor(calledName[0]);
		ArrayList stmtList;
		
		for (String impName : impList) {
			
			String[] names = replaceClassName(calledName, impName);
			stmtList = invokevirtualInst(classFile, names, paramCount, retCount);
			for (int i = 0; i < stmtList.size(); i++) {
				inst.add(stmtList.get(i));
			}
		}
	}
	
	private void invokespecialInst() throws InvalidByteCodeException {
		
		logger.debug("invokespeical instruction");
		String[] calledName = getReferencedName(classFile, absInst);
		int paramCount = PDSMethod.countParam(calledName[2]) + 1;
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		
		if (calledName[0].equals("java/lang/AssertionError")
				&& calledName[1].equals("<init>")) {
			inst.add(gotoStmt(pdsInfo.getLabelAssertError()));
			return;
		}
		
		if (!pdsInfo.isIncludedMethod(calledName)) {
			logger.debug("\tnot included method, paramCount=" + paramCount +
				", retCount=" + retCount);
			if (pdsInfo.checkForNullPointerExceptions()) {
				popPushGTZInst(paramCount, retCount);
			} else {
				popPushInst(paramCount, retCount);
			}
			return;
		}
		
		String calledMethodName = PDSMethod.formatMethodName(calledName);
		String param = paramList(paramCount);
		if (pdsInfo.checkForNullPointerExceptions()) {
			type = IF_INST;
			inst.add(ifStmt(pdsInfo.stackIndex(paramCount - 1) + "==0", 
							 gotoStmt(pdsInfo.getLabelNPE())));
			inst.add(ifStmt("else", popPushStmt(calledMethodName + param, 
					paramCount, retCount)));
		} else {
			popPushInst(calledMethodName + param, paramCount, retCount);
		}
	}
	
	private void invokestaticInst() throws InvalidByteCodeException {
		
		String[] calledName = getReferencedName(classFile, absInst);
		int paramCount = PDSMethod.countParam(calledName[2]);
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		
		if (!pdsInfo.isIncludedMethod(calledName)) {
			if (pdsInfo.checkForNullPointerExceptions()) {
				popPushGTZInst(paramCount, retCount);
			} else {
				popPushInst(paramCount, retCount);
			}
			return;
		}
		
		String calledMethodName = PDSMethod.formatMethodName(calledName);
		String param = paramList(paramCount);
		
		popPushInst(calledMethodName + param, paramCount, retCount);
	}
	
	private void invokevirtualInst() throws InvalidByteCodeException {
		
		String[] calledName = getReferencedName(classFile, absInst);
		int paramCount = PDSMethod.countParam(calledName[2]) + 1;
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		inst = invokevirtualInst(classFile, calledName, paramCount, retCount);
	}
	
	private ArrayList invokevirtualInst(ClassFile cf, String[] calledName,
			int paramCount, int retCount) throws InvalidByteCodeException {
		
		DefaultMutableTreeNode thisNode = pdsInfo.getClassNode(calledName[0]);
		
		if (!pdsInfo.isIncludedMethod(calledName)) {
			
			if ((thisNode != null) && !thisNode.isRoot()) {
				
				DefaultMutableTreeNode superNode 
						= (DefaultMutableTreeNode) thisNode.getParent();
				String superName = (String) superNode.getUserObject();
				ClassFile superCF = pdsInfo.getStoredClassFile(superName);
				String[] superN = replaceClassName(calledName, superName);
				return invokevirtualInst(superCF, superN, paramCount, retCount);
			} else {
				
				if (pdsInfo.checkForNullPointerExceptions()) {
					return popPushGTZStmt(paramCount, retCount);
				} else {
					return popPushStmt(paramCount, retCount);
				}
			}
		} else {
			
			Enumeration e = thisNode.breadthFirstEnumeration();
			DefaultMutableTreeNode subNode;
			ArrayList callInst = new ArrayList();
			String calledMethodName;
			String param = paramList(paramCount);
			String[] subName;
			String heapPtr = pdsInfo.stackIndex(paramCount - 1);
			String objType = pdsInfo.heapIndex(heapPtr);
			type = IF_INST;
			
			while (e.hasMoreElements()) {
				subNode = (DefaultMutableTreeNode) e.nextElement();
				subName = replaceClassName(calledName, 
						(String) subNode.getUserObject());
				if (pdsInfo.isIncludedMethod(subName)) {
					calledMethodName = PDSMethod.formatMethodName(subName);
					callInst.add(ifStmt(objType + "==" + pdsInfo.getClassIndex(subName[0]),
							popPushStmt(calledMethodName + param,
									paramCount, retCount)));
				}
			}
			
			if (pdsInfo.checkForNullPointerExceptions()) {
				callInst.add(ifStmt(heapPtr + "==0", 
						gotoStmt(pdsInfo.getLabelNPE())));
			}
			
			return callInst;
		}
	}
	
	private void iremInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		
		popPushInst(s1 + "-(" + s0 + "*(" + s1 + "/" + s0 + "))", 2, 1);
	}
	
	private void lcmpInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		
		type = IF_INST;
		inst.add(ifStmt(s1 + ">" + s0, s0 + "=1", 2, 1));
		inst.add(ifStmt(s1 + "==" + s0, s0 + "=0", 2, 1));
		inst.add(ifStmt("else", s0 + "=undef", 2, 1));
	}
	
	//TODO: this needs to be tested
	private void ldcInst() throws InvalidByteCodeException {
		
		CPInfo cpInfo = classFile.getConstantPoolEntry(
			immediateByte(), CPInfo.class);
		if (cpInfo instanceof ConstantIntegerInfo) {
			ConstantIntegerInfo ciInfo = (ConstantIntegerInfo) cpInfo;
			pushInst(ciInfo.getInt());
			return;
		}
		
		pushInst("undef");
	}
	
	private void lookupswitchInst() {
		
		LookupSwitchInstruction lsInst = (LookupSwitchInstruction) absInst;
		List moPairs = lsInst.getMatchOffsetPairs();
		int offset = lsInst.getOffset();
		int match;
		int target;
		
		for (int i =  0; i < moPairs.size(); i++) {
			
			match = ((MatchOffsetPair) moPairs.get(i)).getMatch();
			target = offset + ((MatchOffsetPair) moPairs.get(i)).getOffset();
			inst.add(ifStmt(pdsInfo.stackIndex(0) + "==" + match, 
					gotoStmt(methodName + target), 1, 1));
		}
		target = offset + lsInst.getDefaultOffset();
		inst.add(ifStmt("else", gotoStmt(methodName + target), 1, 1));
	}
	
	private void multianewarrayInst() throws InvalidByteCodeException {
		
		int dim = ((MultianewarrayInstruction) absInst).getDimensions();
		String s0 = pdsInfo.stackIndex(0);
		String heapPtr = pdsInfo.getHeapPtrName();
		
		inst.add(s0 + "=" + heapPtr);
		inst.add(pdsInfo.heapIndex(heapPtr) + "=" 
				+ pdsInfo.stackIndex(dim - 1));
		for (int i = dim - 2; i >= 0; i++) {
			
			/* TODO */
		}
	}
	
	private void newInst() throws InvalidByteCodeException {
		
		String newedClassName = getNewedClassName(classFile, absInst);
		
		if (pdsInfo.isIncludedClass(newedClassName)) {
			
			String heapPtr = pdsInfo.getHeapPtrName();
			String classSize = String.valueOf(
					pdsInfo.getClassSize(newedClassName));
			String newHeapPtr = heapPtr + "+" + classSize + "+1";
			String heapUpdate = heapPtr + "=" + newHeapPtr;
			
			ArrayList content = pushStmt(heapPtr);
			content.add(heapUpdate);
			content.add(pdsInfo.heapIndex(heapPtr) 
					+ "=" + pdsInfo.getClassIndex(newedClassName));
			
			if (pdsInfo.checkForHeapOverflow()) {
				type = IF_INST;
				
				inst.add(ifStmt(newHeapPtr + ">" + pdsInfo.getHeapSize(), 
						getOverflowCode()));
				inst.add(ifStmt("else", content));
			} else {
				
				inst = content;
			}
		} else {
			
			pushInst("undef");
		}
	}
	
	/**
	 * Returns statements that handle the case of a Remopla heap overflow.
	 * @return
	 */
	private String getOverflowCode()
	{
		StringBuilder overflow = new StringBuilder("skip;\n\t   ");
		overflow.append(pdsInfo.getLabelHeapOverflow()).append(getLabel());
		overflow.append(":\t");
		overflow.append(gotoStmt(pdsInfo.getLabelHeapOverflow()));
		return overflow.toString();
	}
	
	/*
	 * Allocates (s0 + 1) blocks of memory in the heap.
	 * Note that we use the first block to store the size of array.
	 */
	private void newarrayInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String heapPtr = pdsInfo.getHeapPtrName();
		String newHeapPtr = heapPtr + "+" + s0 + "+1";
		
		ArrayList<String> content = new ArrayList<String>();
		content.add(s0 + "=" + heapPtr);
		content.add(pdsInfo.heapIndex(heapPtr) + "=" + s0);
		content.add(heapPtr + "=" + newHeapPtr);
		
		if (pdsInfo.checkForHeapOverflow()) {
			type = IF_INST;
			inst.add(ifStmt(newHeapPtr + ">" + pdsInfo.getHeapSize(),
					getOverflowCode()));
			inst.add(ifStmt("else", content));
		} else {
			
			inst = content;
		}
	}
	
	private void popInst(String value) {
		
		inst.add(value + "=" + pdsInfo.stackIndex(0));
		popInst(1, 1);
	}
	
	private void popInst(int from, int step) {
		
		for (int i = from; i < maxStack; i++) {
			inst.add(stackEqual(i - step, i));
		}
	}
	
	private void popPushInst(int pop, int push) {
		
		if (pop == 0 && (push == 0)) {
			inst.add("skip");
		} else if ((pop == 0) && (push == 1)) {
			pushInst("undef");
		} else if ((pop == 1) && (push == 1)) {
			inst.add(pdsInfo.stackIndex(0) + "=undef");
		} else if (push == 0) {
			popInst(pop, pop);
		} else {	// if (push == 1)
			inst.add(pdsInfo.stackIndex(0) + "=undef");
			popInst(pop, pop - 1);
		}
	}
	
	private void popPushInst(String action, int pop, int push) {
		
		if ((pop == 0) && (push == 0)) {
			inst.add(action);
		} else if ((pop == 0) && (push == 1)) {
			pushInst(action);
		} else if ((pop == 1) && (push == 1)) {
			inst.add(pdsInfo.stackIndex(0) + "=" + action);
		} else if (push == 0) {
			inst.add(action);
			popInst(pop, pop);
		} else {	// if (push == 1)
			inst.add(pdsInfo.stackIndex(0) + "=" + action);
			popInst(pop, pop - 1);
		}
	}
	
	private void popPushGTZInst(int pop, int push) {
		
		if (pop == 0 && (push == 0)) {
			inst.add("skip");
		} else if ((pop == 0) && (push == 1)) {
			pushInst("!=0");
		} else if ((pop == 1) && (push == 1)) {
			inst.add(pdsInfo.stackIndex(0) + "!=0");
		} else if (push == 0) {
			popInst(pop, pop);
		} else {	// if (push == 1)
			inst.add(pdsInfo.stackIndex(0) + "!=0");
			popInst(pop, pop - 1);
		}
	}
	
	private void pushInst(int value) {
		
		pushInst(String.valueOf(value));
	}
	
	private void pushInst(String value) {
		
		inst.add(pdsInfo.stackIndex(0) + "=" + value);
		pushInst(1, 1);
	}
	
	private void pushInst(int from, int step) {
		
		for (int i = from; i < maxStack; i++) {
			inst.add(stackEqual(i, i - step));
		}
	}
	
	private void putfieldInst() throws InvalidByteCodeException {
		
		String[] fieldName = getReferencedName(classFile, absInst);
		
		if (pdsInfo.isIgnoredField(fieldName)) {
			popInst(2, 2);
			return;
		}
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		DefaultMutableTreeNode thisNode = pdsInfo.getClassNode(fieldName[0]);
		Enumeration e = thisNode.breadthFirstEnumeration();
		DefaultMutableTreeNode subNode;
		String[] subName;
		String formattedSubName;
		String objType = pdsInfo.heapIndex(s1);
		int fieldIndex;
		String h;
		type = IF_INST;
		
		while (e.hasMoreElements()) {
			subNode = (DefaultMutableTreeNode) e.nextElement();
			subName = replaceClassName(fieldName, 
					(String) subNode.getUserObject());
			formattedSubName = PDSField.formatFieldName(subName);
			fieldIndex = pdsInfo.indexOfField(formattedSubName) + 1;
			if (fieldIndex > 0) {
				h = pdsInfo.heapIndex(s1 + "+" + fieldIndex);
				inst.add(ifStmt(
						objType + "==" + pdsInfo.getClassIndex(subName[0]),
						popPushStmt(h + "=" + s0, 2, 0)));
			}
		}
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			inst.add(ifStmt(s1 + "==0", gotoStmt(pdsInfo.getLabelNPE())));
		}
	}
	
	private void putstaticInst() throws InvalidByteCodeException {
		
		String[] fieldName = getReferencedName(classFile, absInst);
		if (pdsInfo.isIgnoredField(fieldName)) {
			popInst(1, 1);
		} else {
			popInst(PDSField.formatFieldName(fieldName));
		}
	}
	
	private void retInst() {
		
		type = IF_INST;
		
		/*TODO*/
	}
	
	private void tableswitchInst() {
		
		TableSwitchInstruction tsInst = (TableSwitchInstruction) absInst;
		int offset = tsInst.getOffset();
		int[] jumpOffsets = tsInst.getJumpOffsets();
		int lowByte = tsInst.getLowByte();
		int highByte = tsInst.getHighByte();
		int target;
		String s0 = pdsInfo.stackIndex(0);
		
		type = IF_INST;
		for (int i = 0; i <= (highByte - lowByte); i++) {
			
			target = offset + jumpOffsets[i];
			inst.add(ifStmt(s0 + "==" + String.valueOf(lowByte + i), 
				gotoStmt(methodName + target), 1, 1));
		}
		target = offset + tsInst.getDefaultOffset();
		inst.add(ifStmt("else", gotoStmt(methodName + target), 1, 1));
	}
	
	/*******************************************************************/
	/**                EXTERNAL UTILITIY FUNCTIONS                    **/
	/*******************************************************************/
	
	/**
	 * Returns name of class in <code>new</code> instruction.
	 *
	 * @param classFile the jclasslib class file.
	 * @param inst the <code>new</code> instruction.
	 * @return the class name.
	 */
	public static String getNewedClassName(ClassFile cFile, 
			AbstractInstruction aInst) throws InvalidByteCodeException {
		
		ImmediateShortInstruction isInst = (ImmediateShortInstruction) aInst;
		ConstantClassInfo ccInfo = 
			(ConstantClassInfo) cFile.getConstantPoolEntry(
				isInst.getImmediateShort(), ConstantClassInfo.class);
				
		return ccInfo.getName();
	}
	
	/**
	 * Returns names of class and method and method's descriptor 
	 * in <code>invokevirtual</code>, <code>invokespecial</code>, 
	 * and <code>invokestatic</code> instruction.
	 *
	 * @param classFile the jclasslib class file.
	 * @param inst the <code>invokevirtual</code>, 
	 *				<code>invokespecial</code>, or <code>invokestatic</code>
	 *				instruction.
	 * @return array of string containing the class name at index 0,
	 *				 method name at index 1, and method descriptor at
	 *				 index 2.
	 */
	public static String[] getReferencedName(ClassFile cFile,
		AbstractInstruction aInst) throws InvalidByteCodeException {
			
		ImmediateShortInstruction isInst = (ImmediateShortInstruction) aInst;
		ConstantReference constRef = (ConstantReference) 
			cFile.getConstantPoolEntry(isInst.getImmediateShort(), 
					ConstantReference.class);
		ConstantClassInfo ccInfo = constRef.getClassInfo();
		ConstantNameAndTypeInfo cnatInfo = constRef.getNameAndTypeInfo();
		
		String[] invokedName = new String[3];
		invokedName[0] = ccInfo.getName();
		invokedName[1] = cnatInfo.getName();
		invokedName[2] = cnatInfo.getDescriptor();
		
		return invokedName;
	}
	
	/**
	 * Tests if the opcode is <code>invokevirtual</code>,
	 * <code>invokespecial</code>, or <code>invokestatic</code>
	 * instruction.
	 *
	 * @param opcode the opcode.
	 * @return <code>true</code> if the opcode is 
	 * <code>invokevirtual</code>, <code>invokespecial</code>, or
	 * <code>invokestatic</code> instruction.
	 */
	public static boolean isInvokeOpcode(int opcode) {
		if ((opcode == Opcodes.OPCODE_INVOKEVIRTUAL)
				|| (opcode == Opcodes.OPCODE_INVOKESPECIAL)
				|| (opcode == Opcodes.OPCODE_INVOKESTATIC))
			return true;
		else
			return false;
	}
	
	/*******************************************************************/
	/**                INTERNAL UTILITIY FUNCTIONS                    **/
	/*******************************************************************/
	
	private String branchTarget() {
		
		BranchInstruction branchInst = (BranchInstruction) absInst;
		int target = branchInst.getOffset() + branchInst.getBranchOffset();
		
		return (target == 0) ? methodName :  methodName + String.valueOf(target);
	}
	
	private String[] calleeInterfaceName() throws InvalidByteCodeException {
		
		InvokeInterfaceInstruction iiInst = (InvokeInterfaceInstruction) absInst;
		ConstantInterfaceMethodrefInfo cimInfo = 
			(ConstantInterfaceMethodrefInfo) classFile.getConstantPoolEntry(
				iiInst.getImmediateShort(), ConstantInterfaceMethodrefInfo.class);
		ConstantClassInfo ccInfo = cimInfo.getClassInfo();
		ConstantNameAndTypeInfo cnatInfo = cimInfo.getNameAndTypeInfo();
		
		String[] out = new String[3];
		out[0] = ccInfo.getName();
		out[1] = cnatInfo.getName();
		out[2] = cnatInfo.getDescriptor();
		
		return out;
	}
	
	private String gotoStmt() {
		
		return "goto " + branchTarget();
	}
	
	private String gotoStmt(String target) {
		
		return "goto " + target;
	}
	
	private ArrayList[] ifStmt(String cond, String stmt, 
			int from, int step) {
		
		ArrayList[] ifthen = ifStmt(cond, stmt);
		for (int i = from; i < maxStack; i++) {
		 ifthen[1].add(stackEqual(i - step, i));
		}
		
		if (ifthen[1].size() == 0) {
			ifthen[1].add("skip");
		}
		
		return ifthen;
	}
	
	private ArrayList[] ifStmt(String cond, String stmt) {
		
		ArrayList[] ifthen = new ArrayList[2];
		
		ifthen[0] = new ArrayList();
		ifthen[0].add(cond);
		
		ifthen[1] = new ArrayList();
		if (stmt != null) {
			ifthen[1].add(stmt);
		}
		
		return ifthen;
	}
	
	private ArrayList[] ifStmt(String cond, ArrayList stmt) {
		
		ArrayList[] ifthen = new ArrayList[2];
		
		ifthen[0] = new ArrayList();
		ifthen[0].add(cond);
		
		ifthen[1] = stmt;
		
		return ifthen;
	}
	
	private int immediateByte() {
		
		return ((ImmediateByteInstruction) absInst).getImmediateByte();
	}
	
	private ArrayList npeStmt(int stackIndex, ArrayList elseStmt) {
		
		ArrayList list = new ArrayList();
		
		type = IF_INST;
		list.add(ifStmt(pdsInfo.stackIndex(stackIndex) + "==0", 
				gotoStmt(pdsInfo.getLabelNPE())));
		list.add(ifStmt("else", elseStmt));
		return list;
	}
	
	private String paramList(int paramCount) {
		
		if (paramCount == 0) {
			return "()";
		}
		
		StringBuffer param = new StringBuffer();
		
		param.append("(");
		param.append(pdsInfo.stackIndex(paramCount - 1));
		for (int i = paramCount - 2; i >= 0; i--) {
			param.append(", ");
			param.append(pdsInfo.stackIndex(i));
		}
		param.append(")");
		
		return param.toString();
	}
	
	private ArrayList popStmt(String value) {
		
		ArrayList list = popStmt(1, 1);
		list.add(0, value + "=" + pdsInfo.stackIndex(0));
		
		return list;
	}
	
	private ArrayList popStmt(int from, int step) {
		
		ArrayList list = new ArrayList();
		
		for (int i = from; i < maxStack; i++) {
			list.add(stackEqual(i - step, i));
		}
		
		return list;
	}
	
	private ArrayList popPushStmt(int pop, int push) {
		
		ArrayList list = new ArrayList();
		
		if ((pop == 0) && (push == 0)) {
			list.add("skip");
		} else if ((pop == 0) && (push == 1)) {
			return pushStmt("undef");
		} else if ((pop == 1) && (push == 1)) {
			list.add(pdsInfo.stackIndex(0) + "=undef");
		} else if (push == 0) {
			return popStmt(pop, pop);
		} else {	// if (push == 1)
			list = popStmt(pop, pop - 1);
			list.add(0, pdsInfo.stackIndex(0) + "=undef");
		}
		
		return list;
	}
	
	private ArrayList popPushStmt(String action, int pop, int push) {
		
		ArrayList list = new ArrayList();
		
		if ((pop == 0) && (push == 0)) {
			list.add(action);
		} else if ((pop == 0) && (push == 1)) {
			list = pushStmt(action);
		} else if ((pop == 1) && (push == 1)) {
			list.add(pdsInfo.stackIndex(0) + "=" + action);
		} else if (push == 0) {
			list = popStmt(pop, pop);
			list.add(0, action);
		} else {	// if (push == 1)
			list = popStmt(pop, pop - 1);
			list.add(0, pdsInfo.stackIndex(0) + "=" + action);
		}
		
		return list;
	}
	
	private ArrayList popPushGTZStmt(int pop, int push) {
		
		ArrayList list = new ArrayList();
		
		if ((pop == 0) && (push == 0)) {
			list.add("skip");
		} else if ((pop == 0) && (push == 1)) {
			list = pushStmt("!=0");
		} else if ((pop == 1) && (push == 1)) {
			list.add(pdsInfo.stackIndex(0) + "!=0");
		} else if (push == 0) {
			list = popStmt(pop, pop);
		} else {	// if (push == 1)
			list = popStmt(pop, pop - 1);
			list.add(0, pdsInfo.stackIndex(0) + "!=0");
		}
		
		return list;
	}
	
	private ArrayList pushStmt(int value) {
		
		return pushStmt(String.valueOf(value));
	}
	
	private ArrayList pushStmt(String value) {
		
		ArrayList list = pushStmt(1, 1);
		list.add(0, pdsInfo.stackIndex(0) + "=" + value);
		
		return list;
	}
	
	private ArrayList pushStmt(int from, int step) {
		
		ArrayList list = new ArrayList();
		for (int i = from; i < maxStack; i++) {
			list.add(stackEqual(i, i - step));
		}
		
		return list;
	}
	
	private String stackArith(String op) {
		
		StringBuffer out = new StringBuffer();
		
		out.append(pdsInfo.stackIndex(0));
		out.append("=");
		out.append(pdsInfo.stackIndex(1));
		out.append(op);
		out.append(pdsInfo.stackIndex(0));
		
		return out.toString();
	}
	
	private String stackEqual(int left, int right) {
		
		StringBuffer out = new StringBuffer();
		
		out.append(pdsInfo.stackIndex(left));
		out.append("=");
		out.append(pdsInfo.stackIndex(right));
		
		return out.toString();
	}
	
}
