package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.HashMap;
import java.util.Map;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public abstract class Coverage 
{

	static {
		System.loadLibrary("mopedcoverage");
	}
	
	private Map<String, CoverageMethod> methodsByLabel =
		new HashMap<String, CoverageMethod>();
	
	private MulticastCoverageListener listenerSupport = new MulticastCoverageListener();
	
	private PDS pds;
	
	private boolean notifyOfReachedLabels = false;
	
	protected Coverage(PDS pds)
	{
		this.pds = pds;
	}
	
	/**
	 * Returns the PDS instance that is used inernally to create the intermediate 
	 * Remopla code.
	 * <p>
	 * You can configure the tester through PDS before you start the tests.
	 * @return
	 */
	public PDS getPDS()
	{
		return pds;
	}
	
	/**
	 * Returns the PDSInfo instance that is used internally to create the 
	 * intermediate Remopla code.
	 * <p>
	 * You can configure the tester through PDS before you start the tests.
	 * @return
	 */
	public PDSInfo getInfo()
	{
		return pds.getInfo();
	}
	
	protected void clearMethodsMap()
	{
		methodsByLabel.clear();
	}
	
	/**
	 * Adds a coverage listener that is notified of the test progress.
	 * @param listener
	 */
	public void addCoverageListener(CoverageListener listener)
	{
		listenerSupport.addCoverageListener(listener);
	}
	
	/**
	 * Removes the coverage listener.
	 * @param listener
	 */
	public void removeCoverageListener(CoverageListener listener)
	{
		listenerSupport.removeCoverageListener(listener);
	}
	
	protected MulticastCoverageListener getListenerSupport()
	{
		return listenerSupport;
	}
	
	protected CoverageMethod[] getMethods(String[] labels)
	{
		for (String label : labels) {
			// register all methods
			getMethod(label);
		}
		return methodsByLabel.values().toArray(new CoverageMethod[0]);
	}
	
	/**
	 * Looks up the method for which the label is used in the remopla output.
	 * @param label
	 * @return <code>null</code> if the method could not be found
	 */
	protected CoverageMethod getMethod(String label) 
	{
		String lookupLabel = CoverageUtils.getMethodLabel(label);
		CoverageMethod method = methodsByLabel.get(lookupLabel);
		if (method == null) {
			method = CoverageUtils.getMethod(pds, label);
			if (method != null) {
				methodsByLabel.put(lookupLabel, method);
			}
		}
		return method;
	}
	
	/**
	 * Needed by the native side.
	 * @return
	 */
	String getAssertErrorLabel()
	{
		return pds.getLabelAssertError();
	}
	
	/**
	 * Needed by the native side.
	 * @return
	 */
	String getHeapOverflowLabel()
	{
		return pds.getLabelHeapOverflow();
	}
	
	public void setEnabledNotifyofReachedLabels(boolean newValue)
	{
		notifyOfReachedLabels = newValue;
	}
	
	public boolean notifyOfReachedLabels()
	{
		return notifyOfReachedLabels;
	}
	
	protected void fireCoveredLabel(String label, int from,  int name, int to) 
	{
		if (notifyOfReachedLabels) {
			getListenerSupport().labelReached(label);
		}

		CoverageMethod method = getMethod(label);
		int pc = CoverageUtils.extractOffset(label);
		if (method != null && pc != -1) {
			method.markPreStarCovered(pc);
			int lineOffset = method.getLineOffset(pc);
			if (method.isFailedAssertionInstruction(pc)) {
				method.markAssertionFailed(pc);
				getListenerSupport().assertErrorReachedFrom(method, lineOffset);
				getListenerSupport().assertErrorReachedFrom(method, lineOffset, 
						new Transition(from, name, to));
			}
			else {
				State state = method.getLineState(lineOffset);
				getListenerSupport().coveredMethodOffset(method, lineOffset,
						state);
			}
		}
	}
	
	protected void fireAssertError(final String label)
	{
		if (notifyOfReachedLabels) {
			getListenerSupport().labelReached(label);
		}
	}
	
	protected void fireHeapOverflow(final String label)
	{
		if (notifyOfReachedLabels) {
			getListenerSupport().labelReached(label);
		}
		
		String methodLabel = label.substring(getHeapOverflowLabel().length());
		
		CoverageMethod method = getMethod(methodLabel);
		int pc = CoverageUtils.extractOffset(methodLabel);
		if (method != null && pc != -1) {
			int lineOffset = method.getLineOffset(pc);
			getListenerSupport().heapOverflow(method, lineOffset);
		}
	}
	
	protected void fireCoverageEnded(CoverageResult result)
	{
		
	}
}
