package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;
import org.gjt.jclasslib.structures.attributes.CodeAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableEntry;

import de.uni_stuttgart.fmi.szs.jmoped.PDSInst;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;

public class CoverageMethod 
{

	public static int ASSERTION_STATEMENTS_COUNT = 4;
	
	private PDSMethod method;
	private LineNumberTableEntry[] lines;
	/**
	 * Array that keeps track of how many instructions are already covered for a line.
	 */
	private int[] coveredInstructions;
	/**
	 * Keeps track of the number of failed assertions per line.
	 * This could be rather sparse, maybe use list or hash map instead.
	 */
	private int[] failedAssertions;

	public static enum State {
		NOT_COVERED,
		PARTIALLY_COVERED,
		COVERED,
		ASSERTION_FAILED,
		HEAP_OVERFLOW
	}
	
	private AnnotatedInstruction[] instructions;
	
	/**
	 * Maps a line offset to the offset in the instructions array, where the instructions
	 * of this line begin.
	 */
	private int[] instructionsOffsets;
	
	private String humanMethodName = null;
	
	private String className = null;
	
	/**
	 * 
	 * @throws NullPointerException if the bytecode of <code>method</code> doesn't
	 * have a line number table
	 */
	public CoverageMethod(PDSMethod method)
	{
		this.method = method;
		CodeAttribute code = (CodeAttribute)method.getMethodInfo().findAttribute(CodeAttribute.class);
		LineNumberTableAttribute table = (LineNumberTableAttribute)
			code.findAttribute(LineNumberTableAttribute.class);
		lines = table.getLineNumberTable();
		coveredInstructions = new int[lines.length];
		failedAssertions = new int[lines.length];
		initialize();
	}

	private void initialize()
	{
		List<AbstractInstruction> pdsInstructions = method.getInstructionList();
		this.instructions = new AnnotatedInstruction[pdsInstructions.size()];
		this.instructionsOffsets = new int[lines.length];
		Iterator<AbstractInstruction> iter = pdsInstructions.iterator();
		int lineOffset = 0;
		
		for (int i = 0; i < instructions.length; i++) {
			AbstractInstruction inst = iter.next();
			instructions[i] = new AnnotatedInstruction(inst, this);
			if (lines.length > lineOffset && lines[lineOffset].getStartPc() == inst.getOffset()) {
				instructionsOffsets[lineOffset] = i;
				lineOffset++;
			}
		}
	}
	
	/**
	 * Returns true if the method is void.
	 * @return
	 */
	public boolean isVoid()
	{
		return PDSMethod.isVoid(method.getFormattedName());
	}
	
	/**
	 * Returns true if this method is the static class initializer method.
	 * @return
	 */
	public boolean isClinit()
	{
		MethodInfo info = method.getMethodInfo();
		try {
			return "<clinit>()V".equals(info.getName() + info.getDescriptor());
		}
		catch (InvalidByteCodeException ivbce) {
		}
		return false;
	}
	
	/**
	 * Returns true if this method is the static class initializer method and it only
	 * initializes the static assertion variables.
	 * @return
	 */
	public boolean clinitIsAssertionOnly()
	{
		if (!isClinit()) {
			return false;
		}
		List<AbstractInstruction> instructions =  method.getInstructionList();
		int staticCount = 0;
		for (AbstractInstruction instruction : instructions) {
			if (instruction.getOpcode() == Opcodes.OPCODE_PUTSTATIC) {
				staticCount++;
				try {
					String[] names = PDSInst.getReferencedName(method.getMethodInfo().getClassFile(), 
							instruction);
					if (!names[1].contains("assertionsDisabled")) {
						staticCount++;
					}
				}
				catch (InvalidByteCodeException ivbce) {
				}
			}
		}
		return staticCount == 1;
	}
	
	/**
	 * Returns true if the line <code>offset</code> contains an assertion.
	 * @param offset
	 * @return
	 */
	public boolean containsAssertion(int offset)
	{
		return getAssertionCount(offset) > 0;
	}
	
	/**
	 * Returns the number off <code>assert</code> statements in this line.
	 * @param offset
	 * @return
	 */
	public int getAssertionCount(int offset)
	{
		int count = 0;
		List<AnnotatedInstruction> instructions = getInstructions(offset);
		for (AnnotatedInstruction i : instructions) {
			AbstractInstruction inst = i.getInstruction();
			if (inst.getOpcode() == Opcodes.OPCODE_GETSTATIC) {
				try {
					String[] names = PDSInst.getReferencedName
						(method.getMethodInfo().getClassFile(), inst);
					if (names[1].contains("assertionsDisabled")) {
						count++;
					}
				}
				catch (InvalidByteCodeException ibce) {
				}
			}
		}
		return count;
	}

	
	/**
	 * Returns true if the method is a constructor.
	 * @return
	 */
	public boolean isConstructor()
	{
		try {
			return method.getMethodInfo().getName().equals("<init>");
		}
		catch (InvalidByteCodeException e) {
		}
		return method.getFormattedName().contains("_init_");
	}
	
	public PDSMethod getMethod()
	{
		return method;
	}
	
	public String getHumanReadableName()
	{
		if (humanMethodName == null) {
			ClassFile classFile = method.getMethodInfo().getClassFile();
			try {
				String className = classFile.getThisClassName();
				int lastIndex = className.lastIndexOf('/');
				if (lastIndex > 0) {
					className = className.substring(lastIndex + 1, className.length());
				}
				humanMethodName = className.replace('$', '.') + "." + method.getName();
			}
			catch (InvalidByteCodeException ive) {
				humanMethodName = method.getFormattedName();
			}
		}
		return humanMethodName;
	}
	
	/**
	 * Returns true if all lines of this method have been fully covered.
	 * @return
	 */
	public boolean isFullyCovered()
	{
		for (int i = 0; i < lines.length; i++) {
			if (getLineState(i) != State.COVERED) {
				return false;
			}
		}
		return true;
	}
	
	@Override
	public String toString() 
	{
		return method.getFormattedName();
	}
	
	void markPostStarCovered(int pc)
	{
		int offset = Arrays.binarySearch(instructions, pc);
		if (offset < 0) {
			throw new IndexOutOfBoundsException("pc not found: " + pc);
		}
		instructions[offset].markPostStarCovered();
	}
	
	void markPreStarCovered(int pc)
	{
		int offset = Arrays.binarySearch(instructions, pc);
		if (offset < 0) {
			throw new IndexOutOfBoundsException("pc not found: " + pc);
		}
		instructions[offset].markPreStarCovered();
	}
	
	void markAssertionFailed(int pc)
	{
		int line = getLineOffset(pc);
		failedAssertions[line]++;
	}
	
	/**
	 * Returns the line offset this bytecode instruction offset belongs to.
	 * 
	 * @throws IndexOutOfBoundsException if <code>pc</code> can't be found in
	 * the line table of this method
	 */
	public int getLineOffset(int pc)
	{
		for (int i = 0; i < lines.length; i++) {
			if (pc >= lines[i].getStartPc()) {
				if (i == lines.length - 1 || lines[i + 1].getStartPc() > pc) {
					return i;
				}
			}
		}
		throw  new IndexOutOfBoundsException("pc wrong " + pc);
	}
	
	/**
	 * Returns the current coverage state for the line at <code>offset</code>.
	 * @param offset
	 * @return
	 */
	public State getLineState(int offset)
	{
		List<AnnotatedInstruction> instructions = getInstructions(offset);
		int coveredCount = 0;
		for (AnnotatedInstruction inst : instructions) {
			coveredCount += inst.getCovered();
		}
		int diff = coveredCount - instructions.size();
		int assertionCount = getAssertionCount(offset);
		// only subtract assertion statements if other statements were reached
		if (assertionCount > 0 && coveredCount > 0) {
			diff += (assertionCount - failedAssertions[offset]) * ASSERTION_STATEMENTS_COUNT;
			diff += failedAssertions[offset];
		}
		if (diff >= 0) {
			return State.COVERED;
		}
		else if (diff == -instructions.size()) {
			return State.NOT_COVERED;
		}
		return State.PARTIALLY_COVERED;
	}
	
	/**
	 * Returns the number of covered instructions for the given line offset.
	 * @param offset
	 * @return
	 */
	public int getCoveredInstructions(int offset)
	{
		return coveredInstructions[offset];
	}
	
	/**
	 * Returns true if the line at <code>offset</code> contains an assertion statement
	 * that has failed in one possible execution path.
	 */
	public boolean containsFailedAssertion(int offset)
	{
		return failedAssertions[offset] > 0;
	}
	
	/**
	 * Returns true if the bytecode instruction at offset <code>pc</code> calls the 
	 * constructor of {@link AssertionError}.
	 */
	public boolean isFailedAssertionInstruction(int pc)
	{
		for (Iterator<AbstractInstruction> i = method.getInstructionList().iterator(); i.hasNext();) {
			AbstractInstruction inst = i.next();
			if (inst.getOffset() == pc) {
				if (inst.getOpcode() == Opcodes.OPCODE_INVOKESPECIAL) {
					try {
						String[] calledName = PDSInst.getReferencedName
						(method.getMethodInfo().getClassFile(), inst);
						if (calledName[0].equals("java/lang/AssertionError")
								&& calledName[1].equals("<init>")) {
							return true;
						}
					}
					catch (InvalidByteCodeException ibce) {
					}
				}
				return false;
			}
		}
		return false;
	}
	
	/**
	 * Returns the number of bytecode instructions that match to the source code
	 * on the line at <code>offset</code>
	 * @param offset the index into the {@link LineNumberTableEntry} array.
	 * 
	 * @throws IndexOutOfBoundsException if the index is negative or too large
	 */
	public int getInstructionCount(int offset)
	{
		int end = offset + 1 < instructionsOffsets.length ? instructionsOffsets[offset + 1] 
		                                                                        : instructions.length;
		return end - instructionsOffsets[offset];
	}
	
	/**
	 * Returns the list of instructions that belong to the line at <code>offset</code>.
	 * @param offset
	 * @return
	 */
	public List<AnnotatedInstruction> getInstructions(int offset)
	{
		int end = offset + 1 < instructionsOffsets.length ? instructionsOffsets[offset + 1] 
		                                                                        : instructions.length;
		List<AnnotatedInstruction> list = new ArrayList<AnnotatedInstruction>();
		
		for (int i = instructionsOffsets[offset]; i < end; i++) { 
			list.add(instructions[i]);
		}
		return list;
	}

	public LineNumberTableEntry[] getLines()
	{
		return lines;
	}

	/**
	 * Returns the line number in the source code corresponding to the offset
	 * in the line number table.
	 * 
	 * @throws IndexOutOfBoundsException if offset is negative or too large
	 */
	public int getLineNumber(int offset)
	{
		return lines[offset].getLineNumber();
	}
	
	/**
	 * Returns true if the line at <code>offset</code> is an effective line, which maps
	 * to a line visible in the source code.
	 */
	public boolean isEffectiveLine(int offset)
	{
		if (isConstructor()) {
			if (offset == 0) {
				return false;
			}
			
		}
		if (isVoid()) {
			if (offset == lines.length - 1 && getInstructionCount(offset) <= 1) {
				return false;
			}
		}
		return true;
	}
	
	public List<String> getUncoveredLabels()
	{
		ArrayList<String> list = new ArrayList<String>();
		for (AnnotatedInstruction inst : instructions) {
			if (!inst.isCovered()) {
				list.add(inst.getLabel());
			}
		}
		return list;
	}
	
	public List<AnnotatedInstruction> getAnnotatedInstructions()
	{
		return Collections.unmodifiableList(Arrays.asList(instructions));
	}
	
	public String getClassName()
	{
		if (className == null) {
			try {
				className = method.getMethodInfo().getClassFile().getThisClassName();
			}
			catch (InvalidByteCodeException ive) {
				// TODO not correct
				className = method.getFormattedName();
			}
		}
		return className;
	}
	
	/**
	 * Returns the number of effective source code lines for this method.
	 * @return
	 */
	public int getLineCount() 
	{
		int count;
		if (isConstructor()) {
			// brace is on same line as last instruction
			if (getInstructionCount(lines.length - 1) > 1) {
				count = lines.length - 1;
			}
			else {
				count = lines.length - 2;
			}
		}
		else if (isVoid()) {
			// brace is on same line as last instruction
			if (getInstructionCount(lines.length - 1) > 1) {
				count = lines.length;
			}
			else {
				count = lines.length - 1;
			}
		}
		else {
			count = lines.length;
		}
		return Math.max(count, 0);
	}
	
}
