package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.io.File;
import java.io.IOException;

import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;

public class CoverageTester extends Coverage
{

	private String className;
	private String methodName;
	private String methodDescriptor;
	private String filterMethodName;
	
	/**
	 * Convenience construtor for {@link #CoverageTester(String, String[], String)
	 * CoverageTester(clazz.getName(), searchPaths, methodName)}.
	 */
	public CoverageTester(Class clazz, String[] searchPaths, String methodName, 
			String methodDescriptor) 
		throws Exception
	{
		this(clazz.getName(), searchPaths, methodName, methodDescriptor);
	}
	
	/**
	 * Constructs a CoverageTester for class <code>className</code>.
	 * <p>
	 * 
	 * @param className should be the fully qualified name of the class as it is returned
	 * by {@link Class#getName()}.
	 * @param searchPaths array of names of directories and jar files that are searched
	 * for the class file of <code>className</code>
	 * @param methodName the method that should be tested for coverage. The format
	 * should match {@link PDSMethod#getFormattedName()}, can be <code>null</code>
	 * @throws Exception
	 */
	public CoverageTester(String className, String[] searchPaths, String methodName,
			String methodDescriptor) 
		throws Exception
	{
		this(new PDS(className, searchPaths, methodName, methodDescriptor),
				className, methodName, methodDescriptor);
	}
	
	public CoverageTester(String className, String[] searchPaths) throws Exception
	{
		this(className, searchPaths, null, null);
	}
	
	public CoverageTester(PDS pds, String className, String methodName, 
			String methodDescriptor)
	{
		super(pds);
		this.className = className;
		this.methodName = methodName;
		this.methodDescriptor = methodDescriptor;
	}
	
	/**
	 * Convenience constructor for {@link #CoverageTester(Class, String[], String),
	 * CoverageTester(clazz, new String[0], null)}.
	 * @param clazz
	 * @throws Exception
	 */
	public CoverageTester(Class clazz) throws Exception
	{
		this(clazz, null, null);
	}
	
	public CoverageTester(Class clazz, String methodName, String methodDescriptor)
		throws Exception
	{
		this(clazz, new String[0], methodName, methodDescriptor);
	}
	
	/**
	 * Returns the name of the class that is to be tested.
	 * @return
	 */
	public String getClassName()
	{
		return className;
	}
	
	/**
	 * May return null.
	 */
	public String getFilterMethodName()
	{
		return filterMethodName;
	}
	
	public void setFilterMethodName(String method)
	{
		filterMethodName = method;
	}
	
	public CoverageMethod[] test()
	{
		return  test(new NullProgressMonitor());
	}
	
	/**
	 * Tests the class' main method for coverage and returns the labels
	 * that have not been reached.
	 * <p>
	 * Listeners are notified of the progress of the test.
	 * 
	 * @param clazz the class must have a correct main method for now
	 * @throws  
	 */
	public CoverageMethod[] test(ProgressMonitor monitor)
	{
		CoverageListener listenerSupport = getListenerSupport();
		clearMethodsMap();
		
		listenerSupport.coverageStarted(getPDS());
		monitor.beginTask("Coverage Analysis", 10000);
		
		File tmpFile = null;
		try {
			tmpFile = CoverageUtils.writeRemopla(getPDS(), className);
		}
		catch (InvalidByteCodeException ibce) {
			listenerSupport.coverageFailed("Invalid byte code", ibce);
			monitor.done();
			return new CoverageMethod[0];
		}
		catch (IOException ie) {
			listenerSupport.coverageFailed("Could not write temporary file", ie);
			monitor.done();
			return new CoverageMethod[0];
		}
		CoverageResult result;
		synchronized (super.getClass()) {
			if (monitor.isCanceled()) {
				monitor.done();
				return new CoverageMethod[0];
			}
			result = analyseCoverage(tmpFile.getAbsolutePath(), filterMethodName, 
					new SubProgressMonitor(10000, monitor));
		}
		
		// TODO does it make sense to say it failed?
		if (monitor.isCanceled()) {
			listenerSupport.coverageFailed("Coverage was cancelled", 
					new InterruptedException());
		}
		else {
			result.setMethods(getMethods(new String[0]));
			result.setPDS(getPDS());
			listenerSupport.coverageEnded(result);
		}
		
		monitor.done();
		return getMethods(new String[0]);
	}
	
	private void unreachedLabels(String[] labels)
	{
		getMethods(labels == null ? new String[0] : labels);
	}

	private native CoverageResult analyseCoverage(String fileName, String method, 
			ProgressMonitor monitor);
	
}
