package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.Arrays;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;

public class JavaMethodArgumentsResolver 
{

	/**
	 * Returns array of method calls on the whole config path.
	 * @param config
	 * @param pds
	 * @return
	 */
	public MethodCall[] resolveMethodCallTrace(IntConfig config, PDS pds)
	{
		ArrayList<MethodCall> calls = new ArrayList<MethodCall>();

		while (config != null) {
			PDSMethod method = pds.getMethodByFormattedName(config.getStackSybmol());
			if (method != null) {
				if (method.isStatic()) {
					MethodCall call = resolveStaticMethodCall(config, method, pds);
					calls.add(call);
				}
			}
			config = config.getNextConfig();
		}
		return calls.toArray(new MethodCall[calls.size()]);
	}
	
	/**
	 * Returns all method calls on the path from config.
	 * 
	 * @return <code>null</code> if method is not called on the config path.
	 */
	public MethodCall[] resolveMethodCalls(IntConfig config, 
			PDSMethod method, PDS pds) 
	{
		if (method.isStatic()) {
			return resolveStaticMethodCalls(config, method, pds);
		}
		else {
			return resolveInstanceMethodCalls(config, method, pds);
		}
	}
	private MethodCall[] resolveStaticMethodCalls(IntConfig config, PDSMethod method,
			PDS pds)
	{
		ArrayList<MethodCall> calls = new ArrayList<MethodCall>(5);
		
		IntConfig callConfig = getNextMethodCallConfig(config, method);
		
		while (callConfig != null) {
			calls.add(resolveStaticMethodCall(callConfig, method, pds));

			callConfig = callConfig.getNextConfig();
			
			if (callConfig != null) {
				callConfig = getNextMethodCallConfig(callConfig, method);
			}
		}
		return calls.toArray(new MethodCall[calls.size()]);
	}
	
	private MethodCall resolveStaticMethodCall(IntConfig callConfig, 
			PDSMethod method, PDS pds) 
	{
		PDSInfo info = pds.getInfo();
		IntVariable[] stackVars = callConfig.getStackVariables();
		IntVariable[] globalVars = callConfig.getGlobalVariables();
		Parameter[] params = DescriptorParser.parse(method.getDescriptor());
		
		ArrayList<MethodArgument> arguments = new ArrayList<MethodArgument>();
		
		for (int i = 0; i < params.length; i++) {
			Parameter param = params[i];
			if (param.isValue()) {
				arguments.add(new IntVariableMethodArgument
						(param, findLocalVariable(stackVars, i, info)));
			}
			else if (param.isValueArray()) {
				arguments.add(new ValueArrayMethodArgument
						(param, getArray(globalVars, findLocalVariable(stackVars, i, info), info)));
			}
			else if (param.isReference()) {
				arguments.add(new ReferenceMethodArgument(param, 
						findLocalVariable(stackVars, i, info), callConfig, pds));
			}
		}
		return new MethodCallImpl(callConfig, 
				arguments.toArray(new MethodArgument[0]), method);
	}
	
	private MethodCall[] resolveInstanceMethodCalls(IntConfig config,
			PDSMethod method, PDS pds) 
	{
		IntConfig callConfig = getNextMethodCallConfig(config, method);
		if (callConfig == null) {
			return null;
		}
		return null;
	}

	public IntConfig getNextMethodCallConfig(IntConfig config, PDSMethod method)
	{
		return config.getConfigAt(method.getFormattedName());
	}
	
	private int[] getArray(IntVariable[] globalVars, IntVariable indexVar, PDSInfo info)
	{
		IntVariable heapVar = findVariable(globalVars, info.getHeapName());
		
		int offset = indexVar.intValue();
		int[] heap = heapVar.arrayValue();
		int size = heap[offset];
		int[] array = new int[size];
		System.arraycopy(heap, offset + 1, array, 0, size);
		return array;
	}
	
	static IntVariable findLocalVariable(IntVariable[] vars, int index, PDSInfo info)
	{
		String name = info.getLocalVarName() + index;
		return findVariable(vars, name);
	}
	
	static IntVariable findVariable(IntVariable[] vars, String name)
	{
		for (IntVariable var : vars) {
			if (var.getName().equals(name)) {
				return var;
			}
		}
		throw new IllegalArgumentException("Not found var " + name);
	}
	
	private static class IntVariableMethodArgument implements MethodArgument
	{
		
		private Parameter param;
		
		private IntVariable variable;

		public IntVariableMethodArgument(Parameter param, IntVariable var)
		{
			this.param = param;
			this.variable = var;
		}
		
		public boolean isValue()
		{
			return true;
		}

		public IntVariable getVariable() 
		{
			return variable;
		}

		public boolean isArray() 
		{
			return false;
		}

		public int[] getArray() 
		{
			return null;
		}

		public boolean isReference() 
		{
			return false;
		}

		public MethodCall getConstructorCall() 
		{
			return null;
		}

		public Parameter getFormalParameter() 
		{
			return param;
		}
		
		@Override
		public String toString() 
		{
			StringBuilder builder = new StringBuilder();
			if (variable.getType() == IntVariable.Type.BOOLEAN) {
				builder.append(variable.booleanValue());
			}
			else {
				builder.append(variable.intValue());
			}
			builder.append(" : ");
			builder.append(param);
			return builder.toString();
		}
		
	}
	
	private static class ValueArrayMethodArgument implements MethodArgument
	{

		private int[] array;
		
		private Parameter param;
		
		public ValueArrayMethodArgument(Parameter param, int[] array)
		{
			this.param = param;
			this.array = array;
		}
		
		public boolean isValue() 
		{
			return false;
		}

		public IntVariable getVariable() 
		{
			return null;
		}

		public boolean isArray() 
		{
			return true;
		}

		public int[] getArray() 
		{
			return array;
		}

		public boolean isReference() 
		{
			return false;
		}

		public MethodCall getConstructorCall() 
		{
			return null;
		}

		public Parameter getFormalParameter() 
		{
			return param;
		}
		
		@Override
		public String toString() 
		{
			StringBuilder builder = new StringBuilder();
			builder.append(Arrays.toString(array));
			builder.append(" : ");
			builder.append(param);
			return builder.toString();
		}
		
	}
	
	private static class MethodCallImpl implements MethodCall
	{
		private IntConfig config;
		
		private MethodArgument[] args;
		
		private PDSMethod method;
		
		public MethodCallImpl(IntConfig config, MethodArgument[] args, 
				PDSMethod method) 
		{
			this.config = config;
			this.args = args;
			this.method = method;
		}

		public MethodArgument[] getArguments() 
		{
			return args;
		}

		public IntConfig getConfig() 
		{
			return config;
		}
		
		public String toString()
		{
			StringBuilder builder = new StringBuilder(config.getStackSybmol());
			builder.append('(');
			for (int i = 0; i < args.length; i++) {
				builder.append(args[i]);
				if (i != args.length - 1) {
					builder.append(", ");
				}
			}
			builder.append(')');
			return builder.toString();
		}

		public PDSMethod getMethod() 
		{
			return method;
		}
	
	}
}
