class RBNode {
	
	boolean isRed;
	int key;
	RBNode left;
	RBNode right;
	RBNode parent;

	/**
	 * Initializes a red-black node.
	 *
	 * @param key		the key of the node.
	 */
	public RBNode(int key) {
		
		isRed = false;
		key = key;
		left = null;
		right = null;
		parent = null;
	}
}

class RBTree {
	
	RBNode root;
	
	/**
	 * Initializes a new red-black tree.
	 */
	RBTree() {
		root = null;
	}
	
	/**
	 * Inserts the red-black node as a leaf 
	 *	into this binary search tree.
	 *
	 * @param z		the red-black node.
	 */
	void TreeInsert(RBNode z) {
		
		RBNode y = null;
		RBNode x = this.root;
		while (x != null) {
			y = x;
			if (z.key < x.key) x = x.left;
			else x = x.right;
		}
		z.parent = y;
		if (y == null) 
			this.root = z;
		else if (z.key < y.key) 
			y.left = z;
		else 
			y.right = z;
	}
	
	/**
	 * Performs left rotation of the red-black tree 
	 * 	at the red-black node.
	 *
	 * @param t		the red-black tree.
	 * @param x		the red-black node.
	 */
	static void LeftRotate(RBTree t, RBNode x) {
		
		RBNode y = x.right;
		x.right = y.left;
		if (y.left != null) 
			y.left.parent = x;
		y.parent = x.parent;
		if (x.parent == null) 
			t.root = y;
		else if (x == x.parent.left) 
			x.parent.left = y;
		else 
			x.parent.right = y;
		y.left = x;
		x.parent = y;
	}
	
	/**
	 * Performs right rotation of the red-black tree 
	 * 	at the red-black node.
	 *
	 * @param t		the red-black tree.
	 * @param y		the red-black node.
	 */
	static void RightRotate(RBTree t, RBNode y) {
		
		RBNode x = y.left;
		y.left = x.right;
		if (x.right != null)
			x.right.parent = y;
		x.parent = y.parent;
		if (y.parent == null)
			t.root = x;
		else if (y == y.parent.left)
			y.parent.left = x;
		else
			y.parent.right = x;
		x.right = y;
		y.parent = x;
	}
	
	/**
	 * Inserts the red-black node into the red-black tree
	 *
	 * @param t		the red-black tree.
	 * @param x		the red-black node.
	 */
	static void RBInsert(RBTree t, RBNode x) {
		
		t.TreeInsert(x);
		x.isRed = true;
		while ((x != t.root) && (x.parent.isRed == true)) {
			if (x.parent.parent == null) {
				x.parent.isRed = false;
			} else if (x.parent == x.parent.parent.left) {
				RBNode y = x.parent.parent.right;
				if ((y != null) && (y.isRed == true)) {
					x.parent.isRed = false;
					y.isRed = false;
					x.parent.parent.isRed = true;
					x = x.parent.parent;
				} else {
					if (x == x.parent.right) {
						x = x.parent;
						LeftRotate(t, x);
					}
					x.parent.isRed = false;	//bug seeded
					x.parent.parent.isRed = true;
					RightRotate(t, x.parent.parent);
				}
			} else {
				RBNode y = x.parent.parent.left;
				if ((y != null) && (y.isRed == true)) {
					x.parent.isRed = false;	//another bug seeded
					y.isRed = false;
					x.parent.parent.isRed = true;
					x = x.parent.parent;
				} else {
					if (x == x.parent.left) {
						x = x.parent;
						RightRotate(t, x);
					}
					x.parent.isRed = false;
					x.parent.parent.isRed = true;
					LeftRotate(t, x.parent.parent);
				}
			}
			t.root.isRed = false;
		}
	}
	
	/**
	 * Main method: creates a tree, inserts nodes, 
	 *	and checks the tree properties.
	 */
	public static void main(String[] args) {
		
		RBTree t = new RBTree();
		
		RBInsert(t, new RBNode(Integer.parseInt(args[0])));
		RBInsert(t, new RBNode(Integer.parseInt(args[0])));
		RBInsert(t, new RBNode(Integer.parseInt(args[0])));
		RBInsert(t, new RBNode(Integer.parseInt(args[0])));
		RBInsert(t, new RBNode(Integer.parseInt(args[0])));
		
		assert(checkProperty3(t.root));
		assert(checkProperty4(t.root));
		ok();
	}
	
	/**
	 * Red-Black Tree Property 3:
	 * 	If a node is red, then both its children are black.
	 *
	 * @param x		the red-black node where the check starts.
	 * @return true if the property 3 holds.
	 */
	static boolean checkProperty3(RBNode x) {
		
		if (x == null) return true;
		if (x.isRed) {
			if (x.left != null && x.left.isRed)
				return false;
			if (x.right != null && x.right.isRed)
				return false;
		}
		return checkProperty3(x.left) && checkProperty3(x.right);
	}
	
	/**
	 * Red-Black Tree Property 4:
	 * 	Every simple path from a node to a descendant leaf 
	 *	contains the same number of blak nodes.
	 *
	 * @param x		the red-black node where the check starts.
	 * @return true if the property 4 holds.
	 */
	static boolean checkProperty4(RBNode x) {
		
		if (blackNodeCount(x, 0) == 0)
			return false;
		return true;
	 
	}
	
	/**
	 * Counts the number of black nodes, comparing its left and right
	 *	returns 0 if unequal.
	 *
	 * @param x			the counting red-black node.
	 * @param count	the number of black nodes couted so far.
	 * @return the number of black nodes, or 0 if its successors
	 *					have different number of black-node counts.
	 */
	static int blackNodeCount(RBNode x, int count) {
		
		if (x == null) return count + 1;
		int countLeft;
		if (x.isRed) {
			countLeft = blackNodeCount(x.left, count);
			if(countLeft != blackNodeCount(x.right, count))
				return 0;
		} else {
			countLeft = blackNodeCount(x.left, count+1);
			if(countLeft != blackNodeCount(x.right, count+1))
				return 0;
		}
		return countLeft;
	}
	
	static void ok() {
		
		while (true) {}
	}
}
